import asyncio
import json
from collections.abc import Callable, Mapping
from enum import Enum
from typing import Any, ParamSpec, TypeGuard, TypeVar, overload

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from hatchet_sdk.context.context import Context, DurableContext
from hatchet_sdk.contracts.v1.workflows_pb2 import Concurrency
from hatchet_sdk.contracts.v1.workflows_pb2 import DefaultFilter as DefaultFilterProto
from hatchet_sdk.utils.timedelta_to_expression import Duration
from hatchet_sdk.utils.typing import (
    AwaitableLike,
    DataclassInstance,
    JSONSerializableMapping,
)

ValidTaskReturnType = BaseModel | Mapping[str, Any] | DataclassInstance | None

R = TypeVar("R", bound=ValidTaskReturnType)
P = ParamSpec("P")


class EmptyModel(BaseModel):
    model_config = ConfigDict(extra="allow", frozen=True)


class StickyStrategy(str, Enum):
    SOFT = "SOFT"
    HARD = "HARD"


class ConcurrencyLimitStrategy(str, Enum):
    CANCEL_IN_PROGRESS = "CANCEL_IN_PROGRESS"
    GROUP_ROUND_ROBIN = "GROUP_ROUND_ROBIN"
    CANCEL_NEWEST = "CANCEL_NEWEST"


class ConcurrencyExpression(BaseModel):
    """
    Defines concurrency limits for a workflow using a CEL expression.
    Args:
        expression (str): CEL expression to determine concurrency grouping. (i.e. "input.user_id")
        max_runs (int): Maximum number of concurrent workflow runs.
        limit_strategy (ConcurrencyLimitStrategy): Strategy for handling limit violations.
    Example:
        ConcurrencyExpression("input.user_id", 5, ConcurrencyLimitStrategy.CANCEL_IN_PROGRESS)
    """

    expression: str
    max_runs: int
    limit_strategy: ConcurrencyLimitStrategy

    def to_proto(self) -> Concurrency:
        return Concurrency(
            expression=self.expression,
            max_runs=self.max_runs,
            limit_strategy=self.limit_strategy,
        )

    @staticmethod
    def from_int(max_runs: int) -> "ConcurrencyExpression":
        return ConcurrencyExpression(
            expression="'constant'",
            max_runs=max_runs,
            limit_strategy=ConcurrencyLimitStrategy.GROUP_ROUND_ROBIN,
        )


TWorkflowInput = TypeVar(
    "TWorkflowInput", bound=BaseModel | DataclassInstance | dict[str, Any]
)


class TaskDefaults(BaseModel):
    schedule_timeout: Duration | None = None
    execution_timeout: Duration | None = None
    priority: int | None = Field(gt=0, lt=4, default=None)
    retries: int | None = None
    backoff_factor: float | None = None
    backoff_max_seconds: int | None = None


class DefaultFilter(BaseModel):
    expression: str
    scope: str
    payload: JSONSerializableMapping = Field(default_factory=dict)

    def to_proto(self) -> DefaultFilterProto:
        payload_json = json.dumps(self.payload, default=str)

        return DefaultFilterProto(
            expression=self.expression,
            scope=self.scope,
            payload=payload_json.encode("utf-8"),
        )


TaskPayloadForInternalUse = (
    type[BaseModel] | type[DataclassInstance] | dict[str, Any] | None
)


class TaskIOValidator:
    def __init__(
        self,
        workflow_input: TypeAdapter[TaskPayloadForInternalUse],
        step_output: TypeAdapter[TaskPayloadForInternalUse],
    ) -> None:
        self.workflow_input = workflow_input
        self.step_output = step_output


class WorkflowConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    name: str
    description: str | None = None
    version: str | None = None
    on_events: list[str] = Field(default_factory=list)
    on_crons: list[str] = Field(default_factory=list)
    sticky: StickyStrategy | None = None
    concurrency: int | ConcurrencyExpression | list[ConcurrencyExpression] | None = None
    input_validator: TypeAdapter[TaskPayloadForInternalUse]
    default_priority: int | None = None

    task_defaults: TaskDefaults = TaskDefaults()
    default_filters: list[DefaultFilter] = Field(default_factory=list)


class StepType(str, Enum):
    DEFAULT = "default"
    ON_FAILURE = "on_failure"
    ON_SUCCESS = "on_success"


AsyncFunc = Callable[[TWorkflowInput, Context], AwaitableLike[R]]
SyncFunc = Callable[[TWorkflowInput, Context], R]
TaskFunc = AsyncFunc[TWorkflowInput, R] | SyncFunc[TWorkflowInput, R]


def is_async_fn(
    fn: TaskFunc[TWorkflowInput, R],
) -> TypeGuard[AsyncFunc[TWorkflowInput, R]]:
    return asyncio.iscoroutinefunction(fn)


def is_sync_fn(
    fn: TaskFunc[TWorkflowInput, R],
) -> TypeGuard[SyncFunc[TWorkflowInput, R]]:
    return not asyncio.iscoroutinefunction(fn)


DurableAsyncFunc = Callable[[TWorkflowInput, DurableContext], AwaitableLike[R]]
DurableSyncFunc = Callable[[TWorkflowInput, DurableContext], R]
DurableTaskFunc = (
    DurableAsyncFunc[TWorkflowInput, R] | DurableSyncFunc[TWorkflowInput, R]
)


def is_durable_async_fn(
    fn: Callable[..., Any],
) -> TypeGuard[DurableAsyncFunc[TWorkflowInput, R]]:
    return asyncio.iscoroutinefunction(fn)


def is_durable_sync_fn(
    fn: DurableTaskFunc[TWorkflowInput, R],
) -> TypeGuard[DurableSyncFunc[TWorkflowInput, R]]:
    return not asyncio.iscoroutinefunction(fn)


_TModel = TypeVar("_TModel", bound=BaseModel)
_TDataclass = TypeVar("_TDataclass", bound=DataclassInstance)
_T = TypeVar("_T")


@overload
def normalize_validator(validator: None) -> type[EmptyModel]: ...


@overload
def normalize_validator(validator: type[_TModel]) -> type[_TModel]: ...


@overload
def normalize_validator(validator: type[_TDataclass]) -> type[_TDataclass]: ...


@overload
def normalize_validator(validator: type[_T]) -> type[_T]: ...


def normalize_validator(validator: object) -> object:
    if validator is None or validator is type(None):
        return EmptyModel

    return validator



# coding: utf-8

# flake8: noqa
"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from hatchet_sdk.clients.rest.models.accept_invite_request import AcceptInviteRequest

# import models into model package
from hatchet_sdk.clients.rest.models.api_error import APIError
from hatchet_sdk.clients.rest.models.api_errors import APIErrors
from hatchet_sdk.clients.rest.models.api_meta import APIMeta
from hatchet_sdk.clients.rest.models.api_meta_auth import APIMetaAuth
from hatchet_sdk.clients.rest.models.api_meta_integration import APIMetaIntegration
from hatchet_sdk.clients.rest.models.api_meta_posthog import APIMetaPosthog
from hatchet_sdk.clients.rest.models.api_resource_meta import APIResourceMeta
from hatchet_sdk.clients.rest.models.api_token import APIToken
from hatchet_sdk.clients.rest.models.bulk_create_event_request import (
    BulkCreateEventRequest,
)
from hatchet_sdk.clients.rest.models.cancel_event_request import CancelEventRequest
from hatchet_sdk.clients.rest.models.concurrency_limit_strategy import (
    ConcurrencyLimitStrategy,
)
from hatchet_sdk.clients.rest.models.concurrency_stat import ConcurrencyStat
from hatchet_sdk.clients.rest.models.create_api_token_request import (
    CreateAPITokenRequest,
)
from hatchet_sdk.clients.rest.models.create_api_token_response import (
    CreateAPITokenResponse,
)
from hatchet_sdk.clients.rest.models.create_cron_workflow_trigger_request import (
    CreateCronWorkflowTriggerRequest,
)
from hatchet_sdk.clients.rest.models.create_event_request import CreateEventRequest
from hatchet_sdk.clients.rest.models.create_pull_request_from_step_run import (
    CreatePullRequestFromStepRun,
)
from hatchet_sdk.clients.rest.models.create_sns_integration_request import (
    CreateSNSIntegrationRequest,
)
from hatchet_sdk.clients.rest.models.create_tenant_alert_email_group_request import (
    CreateTenantAlertEmailGroupRequest,
)
from hatchet_sdk.clients.rest.models.create_tenant_invite_request import (
    CreateTenantInviteRequest,
)
from hatchet_sdk.clients.rest.models.create_tenant_request import CreateTenantRequest
from hatchet_sdk.clients.rest.models.cron_workflows import CronWorkflows
from hatchet_sdk.clients.rest.models.cron_workflows_list import CronWorkflowsList
from hatchet_sdk.clients.rest.models.cron_workflows_method import CronWorkflowsMethod
from hatchet_sdk.clients.rest.models.cron_workflows_order_by_field import (
    CronWorkflowsOrderByField,
)
from hatchet_sdk.clients.rest.models.event import Event
from hatchet_sdk.clients.rest.models.event_data import EventData
from hatchet_sdk.clients.rest.models.event_key_list import EventKeyList
from hatchet_sdk.clients.rest.models.event_list import EventList
from hatchet_sdk.clients.rest.models.event_order_by_direction import (
    EventOrderByDirection,
)
from hatchet_sdk.clients.rest.models.event_order_by_field import EventOrderByField
from hatchet_sdk.clients.rest.models.event_update_cancel200_response import (
    EventUpdateCancel200Response,
)
from hatchet_sdk.clients.rest.models.event_workflow_run_summary import (
    EventWorkflowRunSummary,
)
from hatchet_sdk.clients.rest.models.events import Events
from hatchet_sdk.clients.rest.models.get_step_run_diff_response import (
    GetStepRunDiffResponse,
)
from hatchet_sdk.clients.rest.models.info_get_version200_response import (
    InfoGetVersion200Response,
)
from hatchet_sdk.clients.rest.models.job import Job
from hatchet_sdk.clients.rest.models.job_run import JobRun
from hatchet_sdk.clients.rest.models.job_run_status import JobRunStatus
from hatchet_sdk.clients.rest.models.list_api_tokens_response import (
    ListAPITokensResponse,
)
from hatchet_sdk.clients.rest.models.list_pull_requests_response import (
    ListPullRequestsResponse,
)
from hatchet_sdk.clients.rest.models.list_slack_webhooks import ListSlackWebhooks
from hatchet_sdk.clients.rest.models.list_sns_integrations import ListSNSIntegrations
from hatchet_sdk.clients.rest.models.log_line import LogLine
from hatchet_sdk.clients.rest.models.log_line_level import LogLineLevel
from hatchet_sdk.clients.rest.models.log_line_list import LogLineList
from hatchet_sdk.clients.rest.models.log_line_order_by_direction import (
    LogLineOrderByDirection,
)
from hatchet_sdk.clients.rest.models.log_line_order_by_field import LogLineOrderByField
from hatchet_sdk.clients.rest.models.pagination_response import PaginationResponse
from hatchet_sdk.clients.rest.models.pull_request import PullRequest
from hatchet_sdk.clients.rest.models.pull_request_state import PullRequestState
from hatchet_sdk.clients.rest.models.queue_metrics import QueueMetrics
from hatchet_sdk.clients.rest.models.rate_limit import RateLimit
from hatchet_sdk.clients.rest.models.rate_limit_list import RateLimitList
from hatchet_sdk.clients.rest.models.rate_limit_order_by_direction import (
    RateLimitOrderByDirection,
)
from hatchet_sdk.clients.rest.models.rate_limit_order_by_field import (
    RateLimitOrderByField,
)
from hatchet_sdk.clients.rest.models.recent_step_runs import RecentStepRuns
from hatchet_sdk.clients.rest.models.registered_workflow import RegisteredWorkflow
from hatchet_sdk.clients.rest.models.reject_invite_request import RejectInviteRequest
from hatchet_sdk.clients.rest.models.replay_event_request import ReplayEventRequest
from hatchet_sdk.clients.rest.models.replay_workflow_runs_request import (
    ReplayWorkflowRunsRequest,
)
from hatchet_sdk.clients.rest.models.replay_workflow_runs_response import (
    ReplayWorkflowRunsResponse,
)
from hatchet_sdk.clients.rest.models.rerun_step_run_request import RerunStepRunRequest
from hatchet_sdk.clients.rest.models.schedule_workflow_run_request import (
    ScheduleWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_run_status import ScheduledRunStatus
from hatchet_sdk.clients.rest.models.scheduled_workflows import ScheduledWorkflows
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_filter import (
    ScheduledWorkflowsBulkDeleteFilter,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_request import (
    ScheduledWorkflowsBulkDeleteRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_response import (
    ScheduledWorkflowsBulkDeleteResponse,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_error import (
    ScheduledWorkflowsBulkError,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_item import (
    ScheduledWorkflowsBulkUpdateItem,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_request import (
    ScheduledWorkflowsBulkUpdateRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_response import (
    ScheduledWorkflowsBulkUpdateResponse,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_list import (
    ScheduledWorkflowsList,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_method import (
    ScheduledWorkflowsMethod,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_order_by_field import (
    ScheduledWorkflowsOrderByField,
)
from hatchet_sdk.clients.rest.models.semaphore_slots import SemaphoreSlots
from hatchet_sdk.clients.rest.models.slack_webhook import SlackWebhook
from hatchet_sdk.clients.rest.models.sns_integration import SNSIntegration
from hatchet_sdk.clients.rest.models.step import Step
from hatchet_sdk.clients.rest.models.step_run import StepRun
from hatchet_sdk.clients.rest.models.step_run_archive import StepRunArchive
from hatchet_sdk.clients.rest.models.step_run_archive_list import StepRunArchiveList
from hatchet_sdk.clients.rest.models.step_run_diff import StepRunDiff
from hatchet_sdk.clients.rest.models.step_run_event import StepRunEvent
from hatchet_sdk.clients.rest.models.step_run_event_list import StepRunEventList
from hatchet_sdk.clients.rest.models.step_run_event_reason import StepRunEventReason
from hatchet_sdk.clients.rest.models.step_run_event_severity import StepRunEventSeverity
from hatchet_sdk.clients.rest.models.step_run_status import StepRunStatus
from hatchet_sdk.clients.rest.models.task_stat import TaskStat
from hatchet_sdk.clients.rest.models.task_status_stat import TaskStatusStat
from hatchet_sdk.clients.rest.models.tenant import Tenant
from hatchet_sdk.clients.rest.models.tenant_alert_email_group import (
    TenantAlertEmailGroup,
)
from hatchet_sdk.clients.rest.models.tenant_alert_email_group_list import (
    TenantAlertEmailGroupList,
)
from hatchet_sdk.clients.rest.models.tenant_alerting_settings import (
    TenantAlertingSettings,
)
from hatchet_sdk.clients.rest.models.tenant_environment import TenantEnvironment
from hatchet_sdk.clients.rest.models.tenant_invite import TenantInvite
from hatchet_sdk.clients.rest.models.tenant_invite_list import TenantInviteList
from hatchet_sdk.clients.rest.models.tenant_list import TenantList
from hatchet_sdk.clients.rest.models.tenant_member import TenantMember
from hatchet_sdk.clients.rest.models.tenant_member_list import TenantMemberList
from hatchet_sdk.clients.rest.models.tenant_member_role import TenantMemberRole
from hatchet_sdk.clients.rest.models.tenant_queue_metrics import TenantQueueMetrics
from hatchet_sdk.clients.rest.models.tenant_resource import TenantResource
from hatchet_sdk.clients.rest.models.tenant_resource_limit import TenantResourceLimit
from hatchet_sdk.clients.rest.models.tenant_resource_policy import TenantResourcePolicy
from hatchet_sdk.clients.rest.models.tenant_step_run_queue_metrics import (
    TenantStepRunQueueMetrics,
)
from hatchet_sdk.clients.rest.models.tenant_ui_version import TenantUIVersion
from hatchet_sdk.clients.rest.models.tenant_version import TenantVersion
from hatchet_sdk.clients.rest.models.trigger_workflow_run_request import (
    TriggerWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.update_cron_workflow_trigger_request import (
    UpdateCronWorkflowTriggerRequest,
)
from hatchet_sdk.clients.rest.models.update_scheduled_workflow_run_request import (
    UpdateScheduledWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.update_tenant_alert_email_group_request import (
    UpdateTenantAlertEmailGroupRequest,
)
from hatchet_sdk.clients.rest.models.update_tenant_invite_request import (
    UpdateTenantInviteRequest,
)
from hatchet_sdk.clients.rest.models.update_tenant_member_request import (
    UpdateTenantMemberRequest,
)
from hatchet_sdk.clients.rest.models.update_tenant_request import UpdateTenantRequest
from hatchet_sdk.clients.rest.models.update_worker_request import UpdateWorkerRequest
from hatchet_sdk.clients.rest.models.user import User
from hatchet_sdk.clients.rest.models.user_change_password_request import (
    UserChangePasswordRequest,
)
from hatchet_sdk.clients.rest.models.user_login_request import UserLoginRequest
from hatchet_sdk.clients.rest.models.user_register_request import UserRegisterRequest
from hatchet_sdk.clients.rest.models.user_tenant_memberships_list import (
    UserTenantMembershipsList,
)
from hatchet_sdk.clients.rest.models.user_tenant_public import UserTenantPublic
from hatchet_sdk.clients.rest.models.v1_cancel_task_request import V1CancelTaskRequest
from hatchet_sdk.clients.rest.models.v1_cancelled_tasks import V1CancelledTasks
from hatchet_sdk.clients.rest.models.v1_cel_debug_request import V1CELDebugRequest
from hatchet_sdk.clients.rest.models.v1_cel_debug_response import V1CELDebugResponse
from hatchet_sdk.clients.rest.models.v1_cel_debug_response_status import (
    V1CELDebugResponseStatus,
)
from hatchet_sdk.clients.rest.models.v1_create_filter_request import (
    V1CreateFilterRequest,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request import (
    V1CreateWebhookRequest,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_api_key import (
    V1CreateWebhookRequestAPIKey,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_base import (
    V1CreateWebhookRequestBase,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_basic_auth import (
    V1CreateWebhookRequestBasicAuth,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_hmac import (
    V1CreateWebhookRequestHMAC,
)
from hatchet_sdk.clients.rest.models.v1_dag_children import V1DagChildren
from hatchet_sdk.clients.rest.models.v1_event import V1Event
from hatchet_sdk.clients.rest.models.v1_event_list import V1EventList
from hatchet_sdk.clients.rest.models.v1_event_triggered_run import V1EventTriggeredRun
from hatchet_sdk.clients.rest.models.v1_event_workflow_run_summary import (
    V1EventWorkflowRunSummary,
)
from hatchet_sdk.clients.rest.models.v1_filter import V1Filter
from hatchet_sdk.clients.rest.models.v1_filter_list import V1FilterList
from hatchet_sdk.clients.rest.models.v1_log_line import V1LogLine
from hatchet_sdk.clients.rest.models.v1_log_line_level import V1LogLineLevel
from hatchet_sdk.clients.rest.models.v1_log_line_list import V1LogLineList
from hatchet_sdk.clients.rest.models.v1_replay_task_request import V1ReplayTaskRequest
from hatchet_sdk.clients.rest.models.v1_replayed_tasks import V1ReplayedTasks
from hatchet_sdk.clients.rest.models.v1_task_event import V1TaskEvent
from hatchet_sdk.clients.rest.models.v1_task_event_list import V1TaskEventList
from hatchet_sdk.clients.rest.models.v1_task_event_type import V1TaskEventType
from hatchet_sdk.clients.rest.models.v1_task_filter import V1TaskFilter
from hatchet_sdk.clients.rest.models.v1_task_point_metric import V1TaskPointMetric
from hatchet_sdk.clients.rest.models.v1_task_point_metrics import V1TaskPointMetrics
from hatchet_sdk.clients.rest.models.v1_task_run_metric import V1TaskRunMetric
from hatchet_sdk.clients.rest.models.v1_task_run_status import V1TaskRunStatus
from hatchet_sdk.clients.rest.models.v1_task_status import V1TaskStatus
from hatchet_sdk.clients.rest.models.v1_task_summary import V1TaskSummary
from hatchet_sdk.clients.rest.models.v1_task_summary_list import V1TaskSummaryList
from hatchet_sdk.clients.rest.models.v1_task_timing import V1TaskTiming
from hatchet_sdk.clients.rest.models.v1_task_timing_list import V1TaskTimingList
from hatchet_sdk.clients.rest.models.v1_trigger_workflow_run_request import (
    V1TriggerWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.v1_update_filter_request import (
    V1UpdateFilterRequest,
)
from hatchet_sdk.clients.rest.models.v1_update_webhook_request import (
    V1UpdateWebhookRequest,
)
from hatchet_sdk.clients.rest.models.v1_webhook import V1Webhook
from hatchet_sdk.clients.rest.models.v1_webhook_api_key_auth import V1WebhookAPIKeyAuth
from hatchet_sdk.clients.rest.models.v1_webhook_auth_type import V1WebhookAuthType
from hatchet_sdk.clients.rest.models.v1_webhook_basic_auth import V1WebhookBasicAuth
from hatchet_sdk.clients.rest.models.v1_webhook_hmac_algorithm import (
    V1WebhookHMACAlgorithm,
)
from hatchet_sdk.clients.rest.models.v1_webhook_hmac_auth import V1WebhookHMACAuth
from hatchet_sdk.clients.rest.models.v1_webhook_hmac_encoding import (
    V1WebhookHMACEncoding,
)
from hatchet_sdk.clients.rest.models.v1_webhook_list import V1WebhookList
from hatchet_sdk.clients.rest.models.v1_webhook_source_name import V1WebhookSourceName
from hatchet_sdk.clients.rest.models.v1_workflow_run import V1WorkflowRun
from hatchet_sdk.clients.rest.models.v1_workflow_run_details import V1WorkflowRunDetails
from hatchet_sdk.clients.rest.models.v1_workflow_run_display_name import (
    V1WorkflowRunDisplayName,
)
from hatchet_sdk.clients.rest.models.v1_workflow_run_display_name_list import (
    V1WorkflowRunDisplayNameList,
)
from hatchet_sdk.clients.rest.models.v1_workflow_type import V1WorkflowType
from hatchet_sdk.clients.rest.models.webhook_worker import WebhookWorker
from hatchet_sdk.clients.rest.models.webhook_worker_create_request import (
    WebhookWorkerCreateRequest,
)
from hatchet_sdk.clients.rest.models.webhook_worker_create_response import (
    WebhookWorkerCreateResponse,
)
from hatchet_sdk.clients.rest.models.webhook_worker_created import WebhookWorkerCreated
from hatchet_sdk.clients.rest.models.webhook_worker_list_response import (
    WebhookWorkerListResponse,
)
from hatchet_sdk.clients.rest.models.webhook_worker_request import WebhookWorkerRequest
from hatchet_sdk.clients.rest.models.webhook_worker_request_list_response import (
    WebhookWorkerRequestListResponse,
)
from hatchet_sdk.clients.rest.models.webhook_worker_request_method import (
    WebhookWorkerRequestMethod,
)
from hatchet_sdk.clients.rest.models.worker import Worker
from hatchet_sdk.clients.rest.models.worker_label import WorkerLabel
from hatchet_sdk.clients.rest.models.worker_list import WorkerList
from hatchet_sdk.clients.rest.models.worker_runtime_info import WorkerRuntimeInfo
from hatchet_sdk.clients.rest.models.worker_runtime_sdks import WorkerRuntimeSDKs
from hatchet_sdk.clients.rest.models.worker_type import WorkerType
from hatchet_sdk.clients.rest.models.workflow import Workflow
from hatchet_sdk.clients.rest.models.workflow_concurrency import WorkflowConcurrency
from hatchet_sdk.clients.rest.models.workflow_kind import WorkflowKind
from hatchet_sdk.clients.rest.models.workflow_list import WorkflowList
from hatchet_sdk.clients.rest.models.workflow_metrics import WorkflowMetrics
from hatchet_sdk.clients.rest.models.workflow_run import WorkflowRun
from hatchet_sdk.clients.rest.models.workflow_run_list import WorkflowRunList
from hatchet_sdk.clients.rest.models.workflow_run_order_by_direction import (
    WorkflowRunOrderByDirection,
)
from hatchet_sdk.clients.rest.models.workflow_run_order_by_field import (
    WorkflowRunOrderByField,
)
from hatchet_sdk.clients.rest.models.workflow_run_shape import WorkflowRunShape
from hatchet_sdk.clients.rest.models.workflow_run_shape_item_for_workflow_run_details import (
    WorkflowRunShapeItemForWorkflowRunDetails,
)
from hatchet_sdk.clients.rest.models.workflow_run_status import WorkflowRunStatus
from hatchet_sdk.clients.rest.models.workflow_run_triggered_by import (
    WorkflowRunTriggeredBy,
)
from hatchet_sdk.clients.rest.models.workflow_runs_cancel_request import (
    WorkflowRunsCancelRequest,
)
from hatchet_sdk.clients.rest.models.workflow_runs_metrics import WorkflowRunsMetrics
from hatchet_sdk.clients.rest.models.workflow_runs_metrics_counts import (
    WorkflowRunsMetricsCounts,
)
from hatchet_sdk.clients.rest.models.workflow_tag import WorkflowTag
from hatchet_sdk.clients.rest.models.workflow_trigger_cron_ref import (
    WorkflowTriggerCronRef,
)
from hatchet_sdk.clients.rest.models.workflow_trigger_event_ref import (
    WorkflowTriggerEventRef,
)
from hatchet_sdk.clients.rest.models.workflow_triggers import WorkflowTriggers
from hatchet_sdk.clients.rest.models.workflow_update_request import (
    WorkflowUpdateRequest,
)
from hatchet_sdk.clients.rest.models.workflow_version import WorkflowVersion
from hatchet_sdk.clients.rest.models.workflow_version_definition import (
    WorkflowVersionDefinition,
)
from hatchet_sdk.clients.rest.models.workflow_version_meta import WorkflowVersionMeta
from hatchet_sdk.clients.rest.models.workflow_workers_count import WorkflowWorkersCount



import asyncio
import logging
from collections.abc import Callable
from datetime import timedelta
from typing import Any, Concatenate, ParamSpec, overload

from pydantic import TypeAdapter

from hatchet_sdk import Context, DurableContext
from hatchet_sdk.client import Client
from hatchet_sdk.clients.dispatcher.dispatcher import DispatcherClient
from hatchet_sdk.clients.events import EventClient
from hatchet_sdk.clients.listeners.run_event_listener import RunEventListenerClient
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.features.cel import CELClient
from hatchet_sdk.features.cron import CronClient
from hatchet_sdk.features.filters import FiltersClient
from hatchet_sdk.features.logs import LogsClient
from hatchet_sdk.features.metrics import MetricsClient
from hatchet_sdk.features.rate_limits import RateLimitsClient
from hatchet_sdk.features.runs import RunsClient
from hatchet_sdk.features.scheduled import ScheduledClient
from hatchet_sdk.features.stubs import StubsClient
from hatchet_sdk.features.workers import WorkersClient
from hatchet_sdk.features.workflows import WorkflowsClient
from hatchet_sdk.labels import DesiredWorkerLabel
from hatchet_sdk.logger import logger
from hatchet_sdk.rate_limit import RateLimit
from hatchet_sdk.runnables.types import (
    ConcurrencyExpression,
    DefaultFilter,
    EmptyModel,
    R,
    StickyStrategy,
    TaskDefaults,
    TWorkflowInput,
    WorkflowConfig,
    normalize_validator,
)
from hatchet_sdk.runnables.workflow import BaseWorkflow, Standalone, Workflow
from hatchet_sdk.utils.timedelta_to_expression import Duration
from hatchet_sdk.utils.typing import CoroutineLike
from hatchet_sdk.worker.worker import LifespanFn, Worker

P = ParamSpec("P")


class Hatchet:
    """
    Main client for interacting with the Hatchet SDK.

    This class provides access to various client interfaces and utility methods
    for working with Hatchet workers, workflows, tasks, and our various feature clients.
    """

    def __init__(
        self,
        debug: bool = False,
        client: Client | None = None,
        config: ClientConfig | None = None,
    ):
        if debug:
            logger.setLevel(logging.DEBUG)

        self._client = (
            client if client else Client(config=config or ClientConfig(), debug=debug)
        )

    @property
    def cel(self) -> CELClient:
        """
        The CEL client is a client for interacting with Hatchet's CEL API.
        """
        return self._client.cel

    @property
    def cron(self) -> CronClient:
        """
        The cron client is a client for managing cron workflows within Hatchet.
        """
        return self._client.cron

    @property
    def filters(self) -> FiltersClient:
        """
        The filters client is a client for interacting with Hatchet's filters API.
        """
        return self._client.filters

    @property
    def logs(self) -> LogsClient:
        """
        The logs client is a client for interacting with Hatchet's logs API.
        """
        return self._client.logs

    @property
    def metrics(self) -> MetricsClient:
        """
        The metrics client is a client for reading metrics out of Hatchet's metrics API.
        """
        return self._client.metrics

    @property
    def rate_limits(self) -> RateLimitsClient:
        """
        The rate limits client is a wrapper for Hatchet's gRPC API that makes it easier to work with rate limits in Hatchet.
        """
        return self._client.rate_limits

    @property
    def runs(self) -> RunsClient:
        """
        The runs client is a client for interacting with task and workflow runs within Hatchet.
        """
        return self._client.runs

    @property
    def scheduled(self) -> ScheduledClient:
        """
        The scheduled client is a client for managing scheduled workflows within Hatchet.
        """
        return self._client.scheduled

    @property
    def workers(self) -> WorkersClient:
        """
        The workers client is a client for managing workers programmatically within Hatchet.
        """
        return self._client.workers

    @property
    def workflows(self) -> WorkflowsClient:
        """
        The workflows client is a client for managing workflows programmatically within Hatchet.

        Note that workflows are the declaration, _not_ the individual runs. If you're looking for runs, use the `RunsClient` instead.
        """
        return self._client.workflows

    @property
    def dispatcher(self) -> DispatcherClient:
        return self._client.dispatcher

    @property
    def event(self) -> EventClient:
        """
        The event client, which you can use to push events to Hatchet.
        """
        return self._client.event

    @property
    def listener(self) -> RunEventListenerClient:
        return self._client.listener

    @property
    def stubs(self) -> StubsClient:
        return StubsClient(client=self)

    @property
    def config(self) -> ClientConfig:
        return self._client.config

    @property
    def tenant_id(self) -> str:
        """
        The tenant id you're operating in.
        """
        return self._client.config.tenant_id

    @property
    def namespace(self) -> str:
        """
        The current namespace you're interacting with.
        """
        return self._client.config.namespace

    def worker(
        self,
        name: str,
        slots: int = 100,
        durable_slots: int = 1_000,
        labels: dict[str, str | int] | None = None,
        workflows: list[BaseWorkflow[Any]] | None = None,
        lifespan: LifespanFn | None = None,
    ) -> Worker:
        """
        Create a Hatchet worker on which to run workflows.

        :param name: The name of the worker.

        :param slots: The number of workflow slots on the worker. In other words, the number of concurrent tasks the worker can run at any point in time

        :param durable_slots: The number of durable workflow slots on the worker. In other words, the number of concurrent tasks the worker can run at any point in time that are durable.

        :param labels: A dictionary of labels to assign to the worker. For more details, view examples on affinity and worker labels.

        :param workflows: A list of workflows to register on the worker, as a shorthand for calling `register_workflow` on each or `register_workflows` on all of them.

        :param lifespan: A lifespan function to run on the worker. This function will be called when the worker is started, and can be used to perform any setup or teardown tasks.

        :returns: The created `Worker` object, which exposes an instance method `start` which can be called to start the worker.
        """

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        return Worker(
            name=name,
            slots=slots,
            durable_slots=durable_slots,
            labels=labels,
            config=self._client.config,
            debug=self._client.debug,
            owned_loop=loop is None,
            workflows=workflows,
            lifespan=lifespan,
        )

    @overload
    def workflow(
        self,
        *,
        name: str,
        description: str | None = None,
        input_validator: None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        task_defaults: TaskDefaults = TaskDefaults(),
        default_filters: list[DefaultFilter] | None = None,
    ) -> Workflow[EmptyModel]: ...

    @overload
    def workflow(
        self,
        *,
        name: str,
        description: str | None = None,
        input_validator: type[TWorkflowInput],
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        task_defaults: TaskDefaults = TaskDefaults(),
        default_filters: list[DefaultFilter] | None = None,
    ) -> Workflow[TWorkflowInput]: ...

    def workflow(
        self,
        *,
        name: str,
        description: str | None = None,
        input_validator: type[TWorkflowInput] | None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        task_defaults: TaskDefaults = TaskDefaults(),
        default_filters: list[DefaultFilter] | None = None,
    ) -> Workflow[EmptyModel] | Workflow[TWorkflowInput]:
        """
        Define a Hatchet workflow, which can then declare `task`s and be `run`, `schedule`d, and so on.

        :param name: The name of the workflow.

        :param description: A description for the workflow

        :param input_validator: A Pydantic model to use as a validator for the `input` to the tasks in the workflow. If no validator is provided, defaults to an `EmptyModel` under the hood. The `EmptyModel` is a Pydantic model with no fields specified, and with the `extra` config option set to `"allow"`.

        :param on_events: A list of event triggers for the workflow - events which cause the workflow to be run.

        :param on_crons: A list of cron triggers for the workflow.

        :param version: A version for the workflow

        :param sticky: A sticky strategy for the workflow

        :param default_priority: The priority of the workflow. Higher values will cause this workflow to have priority in scheduling over other, lower priority ones.

        :param concurrency: A concurrency object controlling the concurrency settings for this workflow. If an integer is provided, it is treated as a constant concurrency limit with a `GROUP_ROUND_ROBIN` strategy, which means that only `N` runs of the task may execute at any given time.

        :param task_defaults: A `TaskDefaults` object controlling the default task settings for this workflow.

        :param default_filters: A list of filters to create with the workflow is created. Note that this is a helper to allow you to create filters "declaratively" without needing to make a separate API call once the workflow is created to create them.

        :returns: The created `Workflow` object, which can be used to declare tasks, run the workflow, and so on.
        """

        return Workflow[TWorkflowInput](
            WorkflowConfig(
                name=name,
                version=version,
                description=description,
                on_events=on_events or [],
                on_crons=on_crons or [],
                sticky=sticky,
                concurrency=concurrency,
                input_validator=TypeAdapter(normalize_validator(input_validator)),
                task_defaults=task_defaults,
                default_priority=default_priority,
                default_filters=default_filters or [],
            ),
            self,
        )

    @overload
    def task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> Callable[
        [Callable[Concatenate[EmptyModel, Context, P], R | CoroutineLike[R]]],
        Standalone[EmptyModel, R],
    ]: ...

    @overload
    def task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: type[TWorkflowInput],
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> Callable[
        [Callable[Concatenate[TWorkflowInput, Context, P], R | CoroutineLike[R]]],
        Standalone[TWorkflowInput, R],
    ]: ...

    def task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: type[TWorkflowInput] | None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> (
        Callable[
            [Callable[Concatenate[EmptyModel, Context, P], R | CoroutineLike[R]]],
            Standalone[EmptyModel, R],
        ]
        | Callable[
            [Callable[Concatenate[TWorkflowInput, Context, P], R | CoroutineLike[R]]],
            Standalone[TWorkflowInput, R],
        ]
    ):
        """
        A decorator to transform a function into a standalone Hatchet task that runs as part of a workflow.

        :param name: The name of the task. If not specified, defaults to the name of the function being wrapped by the `task` decorator.

        :param description: An optional description for the task.

        :param input_validator: A Pydantic model to use as a validator for the input to the task. If no validator is provided, defaults to an `EmptyModel`.

        :param on_events: A list of event triggers for the task - events which cause the task to be run.

        :param on_crons: A list of cron triggers for the task.

        :param version: A version for the task.

        :param sticky: A sticky strategy for the task.

        :param default_priority: The priority of the task. Higher values will cause this task to have priority in scheduling.

        :param concurrency: A concurrency object controlling the concurrency settings for this task. If an integer is provided, it is treated as a constant concurrency limit with a `GROUP_ROUND_ROBIN` strategy, which means that only `N` runs of the task may execute at any given time.

        :param schedule_timeout: The maximum time allowed for scheduling the task.

        :param execution_timeout: The maximum time allowed for executing the task.

        :param retries: The number of times to retry the task before failing.

        :param rate_limits: A list of rate limit configurations for the task.

        :param desired_worker_labels: A dictionary of desired worker labels that determine to which worker the task should be assigned.

        :param backoff_factor: The backoff factor for controlling exponential backoff in retries.

        :param backoff_max_seconds: The maximum number of seconds to allow retries with exponential backoff to continue.

        :param default_filters: A list of filters to create with the task is created. Note that this is a helper to allow you to create filters "declaratively" without needing to make a separate API call once the task is created to create them.

        :returns: A decorator which creates a `Standalone` task object.
        """

        def inner(
            func: Callable[
                Concatenate[TWorkflowInput, Context, P], R | CoroutineLike[R]
            ],
        ) -> Standalone[TWorkflowInput, R]:
            inferred_name = name or func.__name__

            workflow = Workflow[TWorkflowInput](
                WorkflowConfig(
                    name=inferred_name,
                    version=version,
                    description=description,
                    on_events=on_events or [],
                    on_crons=on_crons or [],
                    sticky=sticky,
                    default_priority=default_priority,
                    input_validator=TypeAdapter(normalize_validator(input_validator)),
                    default_filters=default_filters or [],
                ),
                self,
            )

            if isinstance(concurrency, list):
                _concurrency = concurrency
            elif isinstance(concurrency, ConcurrencyExpression):
                _concurrency = [concurrency]
            elif isinstance(concurrency, int):
                _concurrency = [ConcurrencyExpression.from_int(concurrency)]
            else:
                _concurrency = []

            task_wrapper = workflow.task(
                name=inferred_name,
                schedule_timeout=schedule_timeout,
                execution_timeout=execution_timeout,
                parents=[],
                retries=retries,
                rate_limits=rate_limits or [],
                desired_worker_labels=desired_worker_labels or {},
                backoff_factor=backoff_factor,
                backoff_max_seconds=backoff_max_seconds,
                concurrency=_concurrency,
            )

            created_task = task_wrapper(func)

            return Standalone[TWorkflowInput, R](
                workflow=workflow,
                task=created_task,
            )

        return inner

    @overload
    def durable_task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> Callable[
        [Callable[Concatenate[EmptyModel, DurableContext, P], R | CoroutineLike[R]]],
        Standalone[EmptyModel, R],
    ]: ...

    @overload
    def durable_task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: type[TWorkflowInput],
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> Callable[
        [
            Callable[
                Concatenate[TWorkflowInput, DurableContext, P], R | CoroutineLike[R]
            ]
        ],
        Standalone[TWorkflowInput, R],
    ]: ...

    def durable_task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        input_validator: type[TWorkflowInput] | None = None,
        on_events: list[str] | None = None,
        on_crons: list[str] | None = None,
        version: str | None = None,
        sticky: StickyStrategy | None = None,
        default_priority: int = 1,
        concurrency: (
            int | ConcurrencyExpression | list[ConcurrencyExpression] | None
        ) = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(seconds=60),
        retries: int = 0,
        rate_limits: list[RateLimit] | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        backoff_factor: float | None = None,
        backoff_max_seconds: int | None = None,
        default_filters: list[DefaultFilter] | None = None,
    ) -> (
        Callable[
            [
                Callable[
                    Concatenate[EmptyModel, DurableContext, P], R | CoroutineLike[R]
                ]
            ],
            Standalone[EmptyModel, R],
        ]
        | Callable[
            [
                Callable[
                    Concatenate[TWorkflowInput, DurableContext, P], R | CoroutineLike[R]
                ]
            ],
            Standalone[TWorkflowInput, R],
        ]
    ):
        """
        A decorator to transform a function into a standalone Hatchet _durable_ task that runs as part of a workflow.

        :param name: The name of the task. If not specified, defaults to the name of the function being wrapped by the `task` decorator.

        :param description: An optional description for the task.

        :param input_validator: A Pydantic model to use as a validator for the input to the task. If no validator is provided, defaults to an `EmptyModel`.

        :param on_events: A list of event triggers for the task - events which cause the task to be run.

        :param on_crons: A list of cron triggers for the task.

        :param version: A version for the task.

        :param sticky: A sticky strategy for the task.

        :param default_priority: The priority of the task. Higher values will cause this task to have priority in scheduling.

        :param concurrency: A concurrency object controlling the concurrency settings for this task. If an integer is provided, it is treated as a constant concurrency limit with a `GROUP_ROUND_ROBIN` strategy, which means that only `N` runs of the task may execute at any given time.

        :param schedule_timeout: The maximum time allowed for scheduling the task.

        :param execution_timeout: The maximum time allowed for executing the task.

        :param retries: The number of times to retry the task before failing.

        :param rate_limits: A list of rate limit configurations for the task.

        :param desired_worker_labels: A dictionary of desired worker labels that determine to which worker the task should be assigned.

        :param backoff_factor: The backoff factor for controlling exponential backoff in retries.

        :param backoff_max_seconds: The maximum number of seconds to allow retries with exponential backoff to continue.

        :param default_filters: A list of filters to create with the task is created. Note that this is a helper to allow you to create filters "declaratively" without needing to make a separate API call once the task is created to create them.

        :returns: A decorator which creates a `Standalone` task object.
        """

        def inner(
            func: Callable[
                Concatenate[TWorkflowInput, DurableContext, P], R | CoroutineLike[R]
            ],
        ) -> Standalone[TWorkflowInput, R]:
            inferred_name = name or func.__name__
            workflow = Workflow[TWorkflowInput](
                WorkflowConfig(
                    name=inferred_name,
                    version=version,
                    description=description,
                    on_events=on_events or [],
                    on_crons=on_crons or [],
                    sticky=sticky,
                    input_validator=TypeAdapter(normalize_validator(input_validator)),
                    default_priority=default_priority,
                    default_filters=default_filters or [],
                ),
                self,
            )

            if isinstance(concurrency, list):
                _concurrency = concurrency
            elif isinstance(concurrency, ConcurrencyExpression):
                _concurrency = [concurrency]
            elif isinstance(concurrency, int):
                _concurrency = [ConcurrencyExpression.from_int(concurrency)]
            else:
                _concurrency = []

            task_wrapper = workflow.durable_task(
                name=inferred_name,
                schedule_timeout=schedule_timeout,
                execution_timeout=execution_timeout,
                parents=[],
                retries=retries,
                rate_limits=rate_limits or [],
                desired_worker_labels=desired_worker_labels or {},
                backoff_factor=backoff_factor,
                backoff_max_seconds=backoff_max_seconds,
                concurrency=_concurrency,
            )

            return Standalone[TWorkflowInput, R](
                workflow=workflow,
                task=task_wrapper(func),
            )

        return inner



from datetime import timedelta

HOUR = 3600
MINUTE = 60

Duration = timedelta | str


def timedelta_to_expr(td: Duration) -> str:
    if isinstance(td, str):
        return td

    ## `total_seconds` gives the entire duration,
    ## while `seconds` gives the seconds component of the timedelta
    ## e.g. 1 day and 1 second would give 86401 total seconds but 1 second
    seconds = int(td.total_seconds())

    ## IMPORTANT: We only support hours, minutes, and seconds on the engine
    if seconds % HOUR == 0:
        return f"{seconds // HOUR}h"
    if seconds % MINUTE == 0:
        return f"{seconds // MINUTE}m"
    return f"{seconds}s"



import asyncio
import multiprocessing
import multiprocessing.context
import os
import re
import signal
import sys
from collections.abc import AsyncGenerator, Callable
from contextlib import AsyncExitStack, asynccontextmanager, suppress
from dataclasses import dataclass, field
from enum import Enum
from multiprocessing import Queue
from multiprocessing.process import BaseProcess
from types import FrameType
from typing import Any, TypeVar
from warnings import warn

from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from prometheus_client import Gauge, generate_latest
from pydantic import BaseModel

from hatchet_sdk.client import Client
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.contracts.v1.workflows_pb2 import CreateWorkflowVersionRequest
from hatchet_sdk.exceptions import LifespanSetupError, LoopAlreadyRunningError
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import Action
from hatchet_sdk.runnables.contextvars import task_count
from hatchet_sdk.runnables.task import Task
from hatchet_sdk.runnables.workflow import BaseWorkflow
from hatchet_sdk.utils.typing import STOP_LOOP_TYPE
from hatchet_sdk.worker.action_listener_process import (
    ActionEvent,
    worker_action_listener_process,
)
from hatchet_sdk.worker.runner.run_loop_manager import WorkerActionRunLoopManager

T = TypeVar("T")


class WorkerStatus(Enum):
    INITIALIZED = 1
    STARTING = 2
    HEALTHY = 3
    UNHEALTHY = 4


@dataclass
class WorkerStartOptions:
    loop: asyncio.AbstractEventLoop | None = field(default=None)


class HealthCheckResponse(BaseModel):
    status: str
    name: str
    slots: int
    actions: list[str]
    labels: dict[str, str | int]
    python_version: str


LifespanGenerator = AsyncGenerator[Any, Any]
LifespanFn = Callable[[], LifespanGenerator]


@asynccontextmanager
async def _create_async_context_manager(
    gen: LifespanGenerator,
) -> AsyncGenerator[None, None]:
    try:
        yield
    finally:
        with suppress(StopAsyncIteration):
            await anext(gen)


class Worker:
    def __init__(
        self,
        name: str,
        config: ClientConfig,
        slots: int,
        durable_slots: int,
        labels: dict[str, str | int] | None = None,
        debug: bool = False,
        owned_loop: bool = True,
        handle_kill: bool = True,
        workflows: list[BaseWorkflow[Any]] | None = None,
        lifespan: LifespanFn | None = None,
    ) -> None:
        self.config = config
        self.name = self.config.apply_namespace(name)
        self.slots = slots
        self.durable_slots = durable_slots
        self.debug = debug
        self.labels = labels or {}
        self.handle_kill = handle_kill
        self.owned_loop = owned_loop

        self.action_registry: dict[str, Task[Any, Any]] = {}
        self.durable_action_registry: dict[str, Task[Any, Any]] = {}

        self.killing: bool = False
        self._status: WorkerStatus = WorkerStatus.INITIALIZED

        self.action_listener_process: BaseProcess | None = None
        self.durable_action_listener_process: BaseProcess | None = None

        self.action_listener_health_check: asyncio.Task[None]

        self.action_runner: WorkerActionRunLoopManager | None = None
        self.durable_action_runner: WorkerActionRunLoopManager | None = None

        self.ctx = multiprocessing.get_context("spawn")

        self.action_queue: Queue[Action | STOP_LOOP_TYPE] = self.ctx.Queue()
        self.event_queue: Queue[ActionEvent] = self.ctx.Queue()

        self.durable_action_queue: Queue[Action | STOP_LOOP_TYPE] = self.ctx.Queue()
        self.durable_event_queue: Queue[ActionEvent] = self.ctx.Queue()

        self.loop: asyncio.AbstractEventLoop | None = None

        self.client = Client(config=self.config, debug=self.debug)

        self._setup_signal_handlers()

        self.worker_status_gauge = Gauge(
            "hatchet_worker_status_" + re.sub(r"\W+", "", name),
            "Current status of the Hatchet worker",
        )

        self.has_any_durable = False
        self.has_any_non_durable = False

        self.lifespan = lifespan
        self.lifespan_stack: AsyncExitStack | None = None

        self.register_workflows(workflows or [])

    def register_workflow_from_opts(self, opts: CreateWorkflowVersionRequest) -> None:
        try:
            self.client.admin.put_workflow(opts)
        except Exception:
            logger.exception(f"failed to register workflow: {opts.name}")
            sys.exit(1)

    def register_workflow(self, workflow: BaseWorkflow[Any]) -> None:
        if not workflow.tasks:
            raise ValueError(
                "workflow must have at least one task registered before registering"
            )

        try:
            self.client.admin.put_workflow(workflow.to_proto())
        except Exception:
            logger.exception(f"failed to register workflow: {workflow.name}")
            sys.exit(1)

        for step in workflow.tasks:
            action_name = workflow._create_action_name(step)

            if step.is_durable:
                self.has_any_durable = True
                self.durable_action_registry[action_name] = step
            else:
                self.has_any_non_durable = True
                self.action_registry[action_name] = step

    def register_workflows(self, workflows: list[BaseWorkflow[Any]]) -> None:
        for workflow in workflows:
            self.register_workflow(workflow)

    @property
    def status(self) -> WorkerStatus:
        return self._status

    def _setup_loop(self) -> None:
        try:
            asyncio.get_running_loop()
            raise LoopAlreadyRunningError(
                "An event loop is already running. This worker requires its own dedicated event loop. "
                "Make sure you're not using asyncio.run() or other loop-creating functions in the main thread."
            )
        except RuntimeError:
            pass

        logger.debug("creating new event loop")
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    async def _health_check_handler(self, request: Request) -> Response:
        response = HealthCheckResponse(
            status=self.status.name,
            name=self.name,
            slots=self.slots or 0,
            actions=list(self.action_registry.keys()),
            labels=self.labels,
            python_version=sys.version,
        ).model_dump()

        return web.json_response(response)

    async def _metrics_handler(self, request: Request) -> Response:
        self.worker_status_gauge.set(1 if self.status == WorkerStatus.HEALTHY else 0)

        return web.Response(body=generate_latest(), content_type="text/plain")

    async def _start_health_server(self) -> None:
        port = self.config.healthcheck.port

        app = web.Application()
        app.add_routes(
            [
                web.get("/health", self._health_check_handler),
                web.get("/metrics", self._metrics_handler),
            ]
        )

        runner = web.AppRunner(app)

        try:
            await runner.setup()
            await web.TCPSite(runner, "0.0.0.0", port).start()
        except Exception:
            logger.exception("failed to start healthcheck server")
            return

        logger.info(f"healthcheck server running on port {port}")

    def start(self, options: WorkerStartOptions = WorkerStartOptions()) -> None:
        if not (self.action_registry or self.durable_action_registry):
            raise ValueError(
                "no actions registered, register workflows before starting worker"
            )

        if options.loop is not None:
            warn(
                "Passing a custom event loop is deprecated and will be removed in the future. This option no longer has any effect",
                DeprecationWarning,
                stacklevel=1,
            )

        self._setup_loop()

        if not self.loop:
            raise RuntimeError("event loop not set, cannot start worker")

        asyncio.run_coroutine_threadsafe(self._aio_start(), self.loop)

        # start the loop and wait until its closed
        self.loop.run_forever()

        if self.handle_kill:
            sys.exit(0)

    async def _aio_start(self) -> None:
        main_pid = os.getpid()

        logger.info("------------------------------------------")
        logger.info("STARTING HATCHET...")
        logger.debug(f"worker runtime starting on PID: {main_pid}")

        self._status = WorkerStatus.STARTING

        if (
            len(self.action_registry.keys()) == 0
            and len(self.durable_action_registry.keys()) == 0
        ):
            raise ValueError(
                "no actions registered, register workflows or actions before starting worker"
            )

        lifespan_context = None
        if self.lifespan:
            try:
                lifespan_context = await self._setup_lifespan()
            except LifespanSetupError as e:
                logger.exception("lifespan setup failed")
                if self.loop:
                    self.loop.stop()
                raise e

        if self.config.healthcheck.enabled:
            await self._start_health_server()

        if self.has_any_non_durable:
            self.action_listener_process = self._start_action_listener(is_durable=False)
            self.action_runner = self._run_action_runner(
                is_durable=False, lifespan_context=lifespan_context
            )

        if self.has_any_durable:
            self.durable_action_listener_process = self._start_action_listener(
                is_durable=True
            )
            self.durable_action_runner = self._run_action_runner(
                is_durable=True, lifespan_context=lifespan_context
            )

        if self.loop:
            self.action_listener_health_check = self.loop.create_task(
                self._check_listener_health()
            )

            await self.action_listener_health_check

    def _run_action_runner(
        self, is_durable: bool, lifespan_context: Any | None
    ) -> WorkerActionRunLoopManager:
        # Retrieve the shared queue
        if self.loop:
            return WorkerActionRunLoopManager(
                self.name + ("_durable" if is_durable else ""),
                self.durable_action_registry if is_durable else self.action_registry,
                self.durable_slots if is_durable else self.slots,
                self.config,
                self.durable_action_queue if is_durable else self.action_queue,
                self.durable_event_queue if is_durable else self.event_queue,
                self.loop,
                self.handle_kill,
                self.client.debug,
                self.labels,
                lifespan_context,
            )

        raise RuntimeError("event loop not set, cannot start action runner")

    async def _setup_lifespan(self) -> Any:
        if self.lifespan is None:
            return None

        self.lifespan_stack = AsyncExitStack()

        try:
            lifespan_gen = self.lifespan()
            context = await anext(lifespan_gen)
            await self.lifespan_stack.enter_async_context(
                _create_async_context_manager(lifespan_gen)
            )
            return context
        except StopAsyncIteration:
            return None
        except Exception as e:
            raise LifespanSetupError("An error occurred during lifespan setup") from e

    async def _cleanup_lifespan(self) -> None:
        try:
            if self.lifespan_stack is not None:
                await self.lifespan_stack.aclose()
        except Exception as e:
            logger.exception("error during lifespan cleanup")
            raise LifespanSetupError("An error occurred during lifespan cleanup") from e

    def _start_action_listener(
        self, is_durable: bool
    ) -> multiprocessing.context.SpawnProcess:
        try:
            process = self.ctx.Process(
                target=worker_action_listener_process,
                args=(
                    self.name + ("_durable" if is_durable else ""),
                    (
                        list(self.durable_action_registry.keys())
                        if is_durable
                        else list(self.action_registry.keys())
                    ),
                    self.durable_slots if is_durable else self.slots,
                    self.config,
                    self.durable_action_queue if is_durable else self.action_queue,
                    self.durable_event_queue if is_durable else self.event_queue,
                    self.handle_kill,
                    self.client.debug,
                    self.labels,
                ),
            )
            process.start()
            logger.debug(f"action listener starting on PID: {process.pid}")

            return process
        except Exception:
            logger.exception("failed to start action listener")
            sys.exit(1)

    async def _check_listener_health(self) -> None:
        logger.debug("starting action listener health check...")
        try:
            while not self.killing:
                if (
                    not self.action_listener_process
                    and not self.durable_action_listener_process
                ) or (
                    self.action_listener_process
                    and self.durable_action_listener_process
                    and not self.action_listener_process.is_alive()
                    and not self.durable_action_listener_process.is_alive()
                ):
                    logger.debug("child action listener process killed...")
                    self._status = WorkerStatus.UNHEALTHY
                    if self.loop:
                        self.loop.create_task(self.exit_gracefully())
                    break

                if (
                    self.config.terminate_worker_after_num_tasks
                    and task_count.value >= self.config.terminate_worker_after_num_tasks
                ):
                    if self.loop:
                        self.loop.create_task(self.exit_gracefully())
                    break

                self._status = WorkerStatus.HEALTHY
                await asyncio.sleep(1)
        except Exception:
            logger.exception("error checking listener health")

    def _setup_signal_handlers(self) -> None:
        signal.signal(
            signal.SIGTERM,
            (
                self._handle_force_quit_signal
                if self.config.force_shutdown_on_shutdown_signal
                else self._handle_exit_signal
            ),
        )
        signal.signal(
            signal.SIGINT,
            (
                self._handle_force_quit_signal
                if self.config.force_shutdown_on_shutdown_signal
                else self._handle_exit_signal
            ),
        )
        signal.signal(signal.SIGQUIT, self._handle_force_quit_signal)

    def _handle_exit_signal(self, signum: int, frame: FrameType | None) -> None:
        sig_name = "SIGTERM" if signum == signal.SIGTERM else "SIGINT"
        logger.info(f"received signal {sig_name}...")
        if self.loop:
            self.loop.create_task(self.exit_gracefully())

    def _handle_force_quit_signal(self, signum: int, frame: FrameType | None) -> None:
        signal_received = signal.Signals(signum).name
        logger.info(f"received {signal_received}...")
        if self.loop:
            self.loop.create_task(self._exit_forcefully())

    async def _close(self) -> None:
        logger.info(f"closing worker '{self.name}'...")
        self.killing = True

        if self.action_runner is not None:
            self.action_runner.cleanup()

        if self.durable_action_runner is not None:
            self.durable_action_runner.cleanup()

        await self.action_listener_health_check

    async def exit_gracefully(self) -> None:
        logger.debug(f"gracefully stopping worker: {self.name}")

        if self.killing:
            return await self._exit_forcefully()

        self.killing = True

        if self.action_runner:
            await self.action_runner.wait_for_tasks()
            await self.action_runner.exit_gracefully()

        if self.durable_action_runner:
            await self.durable_action_runner.wait_for_tasks()
            await self.durable_action_runner.exit_gracefully()

        if self.action_listener_process and self.action_listener_process.is_alive():
            self.action_listener_process.kill()

        if (
            self.durable_action_listener_process
            and self.durable_action_listener_process.is_alive()
        ):
            self.durable_action_listener_process.kill()

        try:
            await self._cleanup_lifespan()
        except LifespanSetupError:
            logger.exception("lifespan cleanup failed")

        await self._close()
        if self.loop and self.owned_loop:
            self.loop.stop()

        logger.info("👋")

    async def _exit_forcefully(self) -> None:
        self.killing = True

        logger.debug(f"forcefully stopping worker: {self.name}")

        await self._close()

        if self.action_listener_process:
            self.action_listener_process.kill()

        if self.durable_action_listener_process:
            self.durable_action_listener_process.kill()

        logger.info("👋")
        sys.exit(1)



import asyncio

from hatchet_sdk.clients.rest.api.worker_api import WorkerApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.update_worker_request import UpdateWorkerRequest
from hatchet_sdk.clients.rest.models.worker import Worker
from hatchet_sdk.clients.rest.models.worker_list import WorkerList
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import BaseRestClient


class WorkersClient(BaseRestClient):
    """
    The workers client is a client for managing workers programmatically within Hatchet.
    """

    def _wa(self, client: ApiClient) -> WorkerApi:
        return WorkerApi(client)

    def get(self, worker_id: str) -> Worker:
        """
        Get a worker by its ID.

        :param worker_id: The ID of the worker to retrieve.
        :return: The worker.
        """
        with self.client() as client:
            worker_get = tenacity_retry(
                self._wa(client).worker_get, self.client_config.tenacity
            )
            return worker_get(worker_id)

    async def aio_get(self, worker_id: str) -> Worker:
        """
        Get a worker by its ID.

        :param worker_id: The ID of the worker to retrieve.
        :return: The worker.
        """
        return await asyncio.to_thread(self.get, worker_id)

    def list(
        self,
    ) -> WorkerList:
        """
        List all workers in the tenant determined by the client config.

        :return: A list of workers.
        """
        with self.client() as client:
            worker_list = tenacity_retry(
                self._wa(client).worker_list, self.client_config.tenacity
            )
            return worker_list(
                tenant=self.client_config.tenant_id,
            )

    async def aio_list(
        self,
    ) -> WorkerList:
        """
        List all workers in the tenant determined by the client config.

        :return: A list of workers.
        """
        return await asyncio.to_thread(self.list)

    def update(self, worker_id: str, opts: UpdateWorkerRequest) -> Worker:
        """
        Update a worker by its ID.

        :param worker_id: The ID of the worker to update.
        :param opts: The update options.
        :return: The updated worker.
        """
        with self.client() as client:
            return self._wa(client).worker_update(
                worker=worker_id,
                update_worker_request=opts,
            )

    async def aio_update(self, worker_id: str, opts: UpdateWorkerRequest) -> Worker:
        """
        Update a worker by its ID.

        :param worker_id: The ID of the worker to update.
        :param opts: The update options.
        :return: The updated worker.
        """
        return await asyncio.to_thread(self.update, worker_id, opts)



import asyncio

from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import BaseRestClient
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts import workflows_pb2 as v0_workflow_protos
from hatchet_sdk.contracts.v1 import workflows_pb2 as workflow_protos
from hatchet_sdk.contracts.workflows_pb2_grpc import WorkflowServiceStub
from hatchet_sdk.metadata import get_metadata
from hatchet_sdk.rate_limit import RateLimitDuration
from hatchet_sdk.utils.proto_enums import convert_python_enum_to_proto


class RateLimitsClient(BaseRestClient):
    """
    The rate limits client is a wrapper for Hatchet's gRPC API that makes it easier to work with rate limits in Hatchet.
    """

    def put(
        self,
        key: str,
        limit: int,
        duration: RateLimitDuration = RateLimitDuration.SECOND,
    ) -> None:
        """
        Put a rate limit for a given key.

        :param key: The key to set the rate limit for.
        :param limit: The rate limit to set.
        :param duration: The duration of the rate limit.

        :return: None
        """

        duration_proto = convert_python_enum_to_proto(
            duration, workflow_protos.RateLimitDuration
        )

        conn = new_conn(self.client_config, False)
        client = WorkflowServiceStub(conn)
        put_rate_limit = tenacity_retry(
            client.PutRateLimit, self.client_config.tenacity
        )

        put_rate_limit(
            v0_workflow_protos.PutRateLimitRequest(
                key=key,
                limit=limit,
                duration=duration_proto,  # type: ignore[arg-type]
            ),
            metadata=get_metadata(self.client_config.token),
        )

    async def aio_put(
        self,
        key: str,
        limit: int,
        duration: RateLimitDuration = RateLimitDuration.SECOND,
    ) -> None:
        """
        Put a rate limit for a given key.

        :param key: The key to set the rate limit for.
        :param limit: The rate limit to set.
        :param duration: The duration of the rate limit.

        :return: None
        """

        await asyncio.to_thread(self.put, key, limit, duration)



from hatchet_sdk.clients.admin import AdminClient
from hatchet_sdk.clients.dispatcher.dispatcher import DispatcherClient
from hatchet_sdk.clients.events import EventClient
from hatchet_sdk.clients.listeners.run_event_listener import RunEventListenerClient
from hatchet_sdk.clients.listeners.workflow_listener import PooledWorkflowRunListener
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.features.cel import CELClient
from hatchet_sdk.features.cron import CronClient
from hatchet_sdk.features.filters import FiltersClient
from hatchet_sdk.features.logs import LogsClient
from hatchet_sdk.features.metrics import MetricsClient
from hatchet_sdk.features.rate_limits import RateLimitsClient
from hatchet_sdk.features.runs import RunsClient
from hatchet_sdk.features.scheduled import ScheduledClient
from hatchet_sdk.features.tenant import TenantClient
from hatchet_sdk.features.workers import WorkersClient
from hatchet_sdk.features.workflows import WorkflowsClient


class Client:
    def __init__(
        self,
        config: ClientConfig,
        event_client: EventClient | None = None,
        admin_client: AdminClient | None = None,
        dispatcher_client: DispatcherClient | None = None,
        workflow_listener: PooledWorkflowRunListener | None = None,
        debug: bool = False,
    ):
        self.config = config
        self.dispatcher = dispatcher_client or DispatcherClient(config)
        self.event = event_client or EventClient(config)
        self.listener = RunEventListenerClient(config)
        self.workflow_listener = workflow_listener or PooledWorkflowRunListener(config)

        self.log_interceptor = config.logger
        self.debug = debug

        self.cel = CELClient(self.config)
        self.cron = CronClient(self.config)
        self.filters = FiltersClient(self.config)
        self.logs = LogsClient(self.config)
        self.metrics = MetricsClient(self.config)
        self.rate_limits = RateLimitsClient(self.config)
        self.runs = RunsClient(
            config=self.config,
            workflow_run_event_listener=self.listener,
            workflow_run_listener=self.workflow_listener,
        )
        self.scheduled = ScheduledClient(self.config)
        self.tenant = TenantClient(self.config)
        self.workers = WorkersClient(self.config)
        self.workflows = WorkflowsClient(self.config)

        self.admin = admin_client or AdminClient(
            config, self.workflow_listener, self.listener, self.runs
        )



from typing import TYPE_CHECKING, Any, overload

from hatchet_sdk.context.context import Context
from hatchet_sdk.runnables.types import EmptyModel, R, TWorkflowInput
from hatchet_sdk.runnables.workflow import Standalone, Workflow

if TYPE_CHECKING:
    from hatchet_sdk import Hatchet


class StubsClient:
    def __init__(self, client: "Hatchet"):
        self.client = client

    @overload
    def workflow(
        self,
        *,
        name: str,
        input_validator: None = None,
    ) -> Workflow[EmptyModel]: ...

    @overload
    def workflow(
        self,
        *,
        name: str,
        input_validator: type[TWorkflowInput],
    ) -> Workflow[TWorkflowInput]: ...

    def workflow(
        self,
        *,
        name: str,
        input_validator: type[TWorkflowInput] | None = None,
    ) -> Workflow[EmptyModel] | Workflow[TWorkflowInput]:
        return self.client.workflow(name=name, input_validator=input_validator)

    @overload
    def task(
        self,
        *,
        name: str,
        input_validator: None = None,
        output_validator: None = None,
    ) -> Standalone[EmptyModel, EmptyModel]: ...

    @overload
    def task(
        self,
        *,
        name: str,
        input_validator: None = None,
        output_validator: type[R],
    ) -> Standalone[EmptyModel, R]: ...

    @overload
    def task(
        self,
        *,
        name: str,
        input_validator: type[TWorkflowInput],
        output_validator: None = None,
    ) -> Standalone[TWorkflowInput, EmptyModel]: ...

    @overload
    def task(
        self,
        *,
        name: str,
        input_validator: type[TWorkflowInput],
        output_validator: type[R],
    ) -> Standalone[TWorkflowInput, R]: ...

    def task(
        self,
        *,
        name: str,
        input_validator: type[TWorkflowInput] | None = None,
        output_validator: type[R] | None = None,
    ) -> (
        Standalone[EmptyModel, R]
        | Standalone[TWorkflowInput, R]
        | Standalone[EmptyModel, EmptyModel]
        | Standalone[TWorkflowInput, EmptyModel]
    ):
        def mock_func(input: Any, ctx: Context) -> Any:
            raise NotImplementedError(
                "This is a stub function and should not be called directly."
            )

        return_type = output_validator if output_validator is not None else EmptyModel
        mock_func.__annotations__ = {
            "input": Any,
            "ctx": Context,
            "return": return_type,
        }

        return self.client.task(name=name, input_validator=input_validator)(mock_func)



# coding: utf-8

"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    StrictStr,
    ValidationError,
    field_validator,
)
from typing_extensions import Literal, Self

from hatchet_sdk.clients.rest.models.v1_create_webhook_request_api_key import (
    V1CreateWebhookRequestAPIKey,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_basic_auth import (
    V1CreateWebhookRequestBasicAuth,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_hmac import (
    V1CreateWebhookRequestHMAC,
)

V1CREATEWEBHOOKREQUEST_ONE_OF_SCHEMAS = [
    "V1CreateWebhookRequestAPIKey",
    "V1CreateWebhookRequestBasicAuth",
    "V1CreateWebhookRequestHMAC",
]


class V1CreateWebhookRequest(BaseModel):
    """
    V1CreateWebhookRequest
    """

    # data type: V1CreateWebhookRequestBasicAuth
    oneof_schema_1_validator: Optional[V1CreateWebhookRequestBasicAuth] = None
    # data type: V1CreateWebhookRequestAPIKey
    oneof_schema_2_validator: Optional[V1CreateWebhookRequestAPIKey] = None
    # data type: V1CreateWebhookRequestHMAC
    oneof_schema_3_validator: Optional[V1CreateWebhookRequestHMAC] = None
    actual_instance: Optional[
        Union[
            V1CreateWebhookRequestAPIKey,
            V1CreateWebhookRequestBasicAuth,
            V1CreateWebhookRequestHMAC,
        ]
    ] = None
    one_of_schemas: Set[str] = {
        "V1CreateWebhookRequestAPIKey",
        "V1CreateWebhookRequestBasicAuth",
        "V1CreateWebhookRequestHMAC",
    }

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator("actual_instance")
    def actual_instance_must_validate_oneof(cls, v):
        instance = V1CreateWebhookRequest.model_construct()
        error_messages = []
        match = 0
        # validate data type: V1CreateWebhookRequestBasicAuth
        if not isinstance(v, V1CreateWebhookRequestBasicAuth):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `V1CreateWebhookRequestBasicAuth`"
            )
        else:
            match += 1
        # validate data type: V1CreateWebhookRequestAPIKey
        if not isinstance(v, V1CreateWebhookRequestAPIKey):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `V1CreateWebhookRequestAPIKey`"
            )
        else:
            match += 1
        # validate data type: V1CreateWebhookRequestHMAC
        if not isinstance(v, V1CreateWebhookRequestHMAC):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `V1CreateWebhookRequestHMAC`"
            )
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when setting `actual_instance` in V1CreateWebhookRequest with oneOf schemas: V1CreateWebhookRequestAPIKey, V1CreateWebhookRequestBasicAuth, V1CreateWebhookRequestHMAC. Details: "
                + ", ".join(error_messages)
            )
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when setting `actual_instance` in V1CreateWebhookRequest with oneOf schemas: V1CreateWebhookRequestAPIKey, V1CreateWebhookRequestBasicAuth, V1CreateWebhookRequestHMAC. Details: "
                + ", ".join(error_messages)
            )
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Union[str, Dict[str, Any]]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into V1CreateWebhookRequestBasicAuth
        try:
            instance.actual_instance = V1CreateWebhookRequestBasicAuth.from_json(
                json_str
            )
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into V1CreateWebhookRequestAPIKey
        try:
            instance.actual_instance = V1CreateWebhookRequestAPIKey.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into V1CreateWebhookRequestHMAC
        try:
            instance.actual_instance = V1CreateWebhookRequestHMAC.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when deserializing the JSON string into V1CreateWebhookRequest with oneOf schemas: V1CreateWebhookRequestAPIKey, V1CreateWebhookRequestBasicAuth, V1CreateWebhookRequestHMAC. Details: "
                + ", ".join(error_messages)
            )
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into V1CreateWebhookRequest with oneOf schemas: V1CreateWebhookRequestAPIKey, V1CreateWebhookRequestBasicAuth, V1CreateWebhookRequestHMAC. Details: "
                + ", ".join(error_messages)
            )
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(
            self.actual_instance.to_json
        ):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(
        self,
    ) -> Optional[
        Union[
            Dict[str, Any],
            V1CreateWebhookRequestAPIKey,
            V1CreateWebhookRequestBasicAuth,
            V1CreateWebhookRequestHMAC,
        ]
    ]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(
            self.actual_instance.to_dict
        ):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())



import json
from collections.abc import Callable, Collection, Coroutine
from importlib.metadata import version
from typing import Any, cast

from hatchet_sdk.contracts import workflows_pb2 as v0_workflow_protos
from hatchet_sdk.utils.typing import JSONSerializableMapping

try:
    from opentelemetry.context import Context
    from opentelemetry.instrumentation.instrumentor import (  # type: ignore[attr-defined]
        BaseInstrumentor,
    )
    from opentelemetry.instrumentation.utils import unwrap
    from opentelemetry.metrics import MeterProvider, NoOpMeterProvider, get_meter
    from opentelemetry.trace import (
        NoOpTracerProvider,
        SpanKind,
        StatusCode,
        TracerProvider,
        get_tracer,
        get_tracer_provider,
    )
    from opentelemetry.trace.propagation.tracecontext import (
        TraceContextTextMapPropagator,
    )
    from wrapt import wrap_function_wrapper  # type: ignore[import-untyped]
except (RuntimeError, ImportError, ModuleNotFoundError) as e:
    raise ModuleNotFoundError(
        "To use the HatchetInstrumentor, you must install Hatchet's `otel` extra using (e.g.) `pip install hatchet-sdk[otel]`"
    ) from e

import inspect
from datetime import datetime

from google.protobuf import timestamp_pb2

import hatchet_sdk
from hatchet_sdk import ClientConfig
from hatchet_sdk.clients.admin import (
    AdminClient,
    ScheduleTriggerWorkflowOptions,
    TriggerWorkflowOptions,
    WorkflowRunTriggerConfig,
)
from hatchet_sdk.clients.events import (
    BulkPushEventOptions,
    BulkPushEventWithMetadata,
    EventClient,
    PushEventOptions,
)
from hatchet_sdk.contracts.events_pb2 import Event
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import Action
from hatchet_sdk.utils.opentelemetry import OTelAttribute
from hatchet_sdk.worker.runner.runner import Runner
from hatchet_sdk.workflow_run import WorkflowRunRef

hatchet_sdk_version = version("hatchet-sdk")

InstrumentKwargs = TracerProvider | MeterProvider | None

OTEL_TRACEPARENT_KEY = "traceparent"


def create_traceparent() -> str | None:
    logger.warning(
        "as of SDK version 1.11.0, you no longer need to call `create_traceparent` manually. The traceparent will be automatically created by the instrumentor and injected into the metadata of actions and events when appropriate. This method will be removed in a future version.",
    )
    return _create_traceparent()


def _create_traceparent() -> str | None:
    """
    Creates and returns a W3C traceparent header value using OpenTelemetry's context propagation.

    The traceparent header is used to propagate context information across service boundaries
    in distributed tracing systems. It follows the W3C Trace Context specification.

    :returns: A W3C-formatted traceparent header value if successful, None if the context
                    injection fails or no active span exists.\n
                    Example: `00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01`
    """

    carrier: dict[str, str] = {}
    TraceContextTextMapPropagator().inject(carrier)

    return carrier.get("traceparent")


def parse_carrier_from_metadata(
    metadata: JSONSerializableMapping | None,
) -> Context | None:
    logger.warning(
        "as of SDK version 1.11.0, you no longer need to call `parse_carrier_from_metadata` manually. This method will be removed in a future version.",
    )

    return _parse_carrier_from_metadata(metadata)


def _parse_carrier_from_metadata(
    metadata: JSONSerializableMapping | None,
) -> Context | None:
    """
    Parses OpenTelemetry trace context from a metadata dictionary.

    Extracts the trace context from metadata using the W3C Trace Context format,
    specifically looking for the `traceparent` header.

    :param metadata: A dictionary containing metadata key-value pairs,
                     potentially including the `traceparent` header. Can be None.

    :returns: The extracted OpenTelemetry Context object if a valid `traceparent`
              is found in the metadata, otherwise None.

    :Example:

    >>> metadata = {"traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"}
    >>> context = _parse_carrier_from_metadata(metadata)
    """

    if not metadata:
        return None

    traceparent = metadata.get(OTEL_TRACEPARENT_KEY)

    if not traceparent:
        return None

    return TraceContextTextMapPropagator().extract({OTEL_TRACEPARENT_KEY: traceparent})


def inject_traceparent_into_metadata(
    metadata: dict[str, str], traceparent: str | None = None
) -> dict[str, str]:
    logger.warning(
        "as of SDK version 1.11.0, you no longer need to call `inject_traceparent_into_metadata` manually. The traceparent will automatically be injected by the instrumentor. This method will be removed in a future version.",
    )

    return _inject_traceparent_into_metadata(metadata, traceparent)


def _inject_traceparent_into_metadata(
    metadata: dict[str, str], traceparent: str | None = None
) -> dict[str, str]:
    """
    Injects OpenTelemetry `traceparent` into a metadata dictionary.

    Takes a metadata dictionary and an optional `traceparent` string,
    returning a new metadata dictionary with the `traceparent` added under the
    `OTEL_TRACEPARENT_KEY`. If no `traceparent` is provided, it attempts to create one.

    :param metadata: The metadata dictionary to inject the `traceparent` into.

    :param traceparent: The `traceparent` string to inject. If None, attempts to use
                        the current span.

    :returns: A new metadata dictionary containing the original metadata plus
              the injected `traceparent`, if one was available or could be created.

    :Example:

    >>> metadata = {"key": "value"}
    >>> new_metadata = inject_traceparent(metadata, "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01")
    >>> print(new_metadata)
    {"key": "value", "traceparent": "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"}
    """
    if not traceparent:
        traceparent = _create_traceparent()

    if not traceparent:
        return metadata

    return {
        **metadata,
        OTEL_TRACEPARENT_KEY: traceparent,
    }


class HatchetInstrumentor(BaseInstrumentor):  # type: ignore[misc]
    """
    Hatchet OpenTelemetry instrumentor.

    The instrumentor provides an OpenTelemetry integration for Hatchet by setting up
    tracing and metrics collection.

    :param tracer_provider: TracerProvider | None: The OpenTelemetry TracerProvider to use.
            If not provided, the global tracer provider will be used.
    :param meter_provider: MeterProvider | None: The OpenTelemetry MeterProvider to use.
            If not provided, a no-op meter provider will be used.
    :param config: ClientConfig | None: The configuration for the Hatchet client. If not provided,
            a default configuration will be used.
    """

    def __init__(
        self,
        tracer_provider: TracerProvider | None = None,
        meter_provider: MeterProvider | None = None,
        config: ClientConfig | None = None,
    ):
        self.config = config or ClientConfig()

        self.tracer_provider = tracer_provider or get_tracer_provider()
        self.meter_provider = meter_provider or NoOpMeterProvider()

        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return ()

    def _instrument(self, **kwargs: InstrumentKwargs) -> None:
        self._tracer = get_tracer(__name__, hatchet_sdk_version, self.tracer_provider)
        self._meter = get_meter(__name__, hatchet_sdk_version, self.meter_provider)

        wrap_function_wrapper(
            hatchet_sdk,
            "worker.runner.runner.Runner.handle_start_step_run",
            self._wrap_handle_start_step_run,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "worker.runner.runner.Runner.handle_cancel_action",
            self._wrap_handle_cancel_action,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.events.EventClient.push",
            self._wrap_push_event,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.events.EventClient.bulk_push",
            self._wrap_bulk_push_event,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.run_workflow",
            self._wrap_run_workflow,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.aio_run_workflow",
            self._wrap_async_run_workflow,
        )

        ## IMPORTANT: We don't need to instrument the async version of `schedule_workflow`
        ## because it just calls the sync version internally.
        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.schedule_workflow",
            self._wrap_schedule_workflow,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.run_workflows",
            self._wrap_run_workflows,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.aio_run_workflows",
            self._wrap_async_run_workflows,
        )

    def extract_bound_args(
        self,
        wrapped_func: Callable[..., Any],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> list[Any]:
        sig = inspect.signature(wrapped_func)

        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        return list(bound_args.arguments.values())

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_handle_start_step_run(
        self,
        wrapped: Callable[[Action], Coroutine[None, None, Exception | None]],
        instance: Runner,
        args: tuple[Action],
        kwargs: Any,
    ) -> Exception | None:
        params = self.extract_bound_args(wrapped, args, kwargs)

        action = cast(Action, params[0])

        traceparent = _parse_carrier_from_metadata(action.additional_metadata)
        span_name = "hatchet.start_step_run"

        if self.config.otel.include_task_name_in_start_step_run_span_name:
            span_name += f".{action.action_id}"

        with self._tracer.start_as_current_span(
            span_name,
            attributes=action.get_otel_attributes(self.config),
            context=traceparent,
            kind=SpanKind.CONSUMER,
        ) as span:
            result = await wrapped(*args, **kwargs)

            if isinstance(result, Exception):
                span.set_status(StatusCode.ERROR, str(result))

            return result

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_handle_cancel_action(
        self,
        wrapped: Callable[[Action], Coroutine[None, None, Exception | None]],
        instance: Runner,
        args: tuple[Action],
        kwargs: Any,
    ) -> Exception | None:
        action = args[0]

        with self._tracer.start_as_current_span(
            "hatchet.cancel_step_run",
            attributes={
                "hatchet.step_run_id": action.step_run_id,
            },
            kind=SpanKind.CONSUMER,
        ):
            return await wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_push_event(
        self,
        wrapped: Callable[[str, dict[str, Any], PushEventOptions], Event],
        instance: EventClient,
        args: tuple[
            str,
            dict[str, Any],
            PushEventOptions,
        ],
        kwargs: dict[str, str | dict[str, Any] | PushEventOptions],
    ) -> Event:
        params = self.extract_bound_args(wrapped, args, kwargs)

        event_key = cast(str, params[0])
        payload = cast(JSONSerializableMapping, params[1])
        options = cast(
            PushEventOptions,
            params[2] if len(params) > 2 else PushEventOptions(),
        )

        attributes = {
            OTelAttribute.EVENT_KEY: event_key,
            OTelAttribute.ACTION_PAYLOAD: json.dumps(payload, default=str),
            OTelAttribute.ADDITIONAL_METADATA: json.dumps(
                options.additional_metadata, default=str
            ),
            OTelAttribute.NAMESPACE: options.namespace,
            OTelAttribute.PRIORITY: options.priority,
            OTelAttribute.FILTER_SCOPE: options.scope,
        }

        with self._tracer.start_as_current_span(
            "hatchet.push_event",
            attributes={
                f"hatchet.{k.value}": v
                for k, v in attributes.items()
                if v
                and k not in self.config.otel.excluded_attributes
                and v != "{}"
                and v != "[]"
            },
            kind=SpanKind.PRODUCER,
        ):
            options = PushEventOptions(
                **options.model_dump(exclude={"additional_metadata"}),
                additional_metadata=_inject_traceparent_into_metadata(
                    options.additional_metadata,
                ),
            )

            return wrapped(event_key, payload, options)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_bulk_push_event(
        self,
        wrapped: Callable[
            [list[BulkPushEventWithMetadata], BulkPushEventOptions], list[Event]
        ],
        instance: EventClient,
        args: tuple[
            list[BulkPushEventWithMetadata],
            BulkPushEventOptions,
        ],
        kwargs: dict[str, list[BulkPushEventWithMetadata] | BulkPushEventOptions],
    ) -> list[Event]:
        params = self.extract_bound_args(wrapped, args, kwargs)

        bulk_events = cast(list[BulkPushEventWithMetadata], params[0])
        options = cast(BulkPushEventOptions, params[1])

        num_bulk_events = len(bulk_events)
        unique_event_keys = {event.key for event in bulk_events}

        with self._tracer.start_as_current_span(
            "hatchet.bulk_push_event",
            attributes={
                "hatchet.num_events": num_bulk_events,
                "hatchet.unique_event_keys": json.dumps(unique_event_keys, default=str),
            },
            kind=SpanKind.PRODUCER,
        ):
            bulk_events_with_meta = [
                BulkPushEventWithMetadata(
                    **event.model_dump(exclude={"additional_metadata"}),
                    additional_metadata=_inject_traceparent_into_metadata(
                        event.additional_metadata,
                    ),
                )
                for event in bulk_events
            ]

            return wrapped(
                bulk_events_with_meta,
                options,
            )

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_run_workflow(
        self,
        wrapped: Callable[
            [str, JSONSerializableMapping, TriggerWorkflowOptions],
            WorkflowRunRef,
        ],
        instance: AdminClient,
        args: tuple[str, JSONSerializableMapping, TriggerWorkflowOptions],
        kwargs: dict[str, str | JSONSerializableMapping | TriggerWorkflowOptions],
    ) -> WorkflowRunRef:
        params = self.extract_bound_args(wrapped, args, kwargs)

        workflow_name = cast(str, params[0])
        payload = cast(JSONSerializableMapping, params[1])
        options = cast(
            TriggerWorkflowOptions,
            params[2] if len(params) > 2 else TriggerWorkflowOptions(),
        )

        attributes = {
            OTelAttribute.WORKFLOW_NAME: workflow_name,
            OTelAttribute.ACTION_PAYLOAD: json.dumps(payload, default=str),
            OTelAttribute.PARENT_ID: options.parent_id,
            OTelAttribute.PARENT_STEP_RUN_ID: options.parent_step_run_id,
            OTelAttribute.CHILD_INDEX: options.child_index,
            OTelAttribute.CHILD_KEY: options.child_key,
            OTelAttribute.NAMESPACE: options.namespace,
            OTelAttribute.ADDITIONAL_METADATA: json.dumps(
                options.additional_metadata, default=str
            ),
            OTelAttribute.PRIORITY: options.priority,
            OTelAttribute.DESIRED_WORKER_ID: options.desired_worker_id,
            OTelAttribute.STICKY: options.sticky,
            OTelAttribute.KEY: options.key,
        }

        with self._tracer.start_as_current_span(
            "hatchet.run_workflow",
            attributes={
                f"hatchet.{k.value}": v
                for k, v in attributes.items()
                if v
                and k not in self.config.otel.excluded_attributes
                and v != "{}"
                and v != "[]"
            },
            kind=SpanKind.PRODUCER,
        ):
            options = TriggerWorkflowOptions(
                **options.model_dump(exclude={"additional_metadata"}),
                additional_metadata=_inject_traceparent_into_metadata(
                    options.additional_metadata,
                ),
            )

            return wrapped(workflow_name, payload, options)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_async_run_workflow(
        self,
        wrapped: Callable[
            [str, JSONSerializableMapping, TriggerWorkflowOptions],
            Coroutine[None, None, WorkflowRunRef],
        ],
        instance: AdminClient,
        args: tuple[str, JSONSerializableMapping, TriggerWorkflowOptions],
        kwargs: dict[str, str | JSONSerializableMapping | TriggerWorkflowOptions],
    ) -> WorkflowRunRef:
        params = self.extract_bound_args(wrapped, args, kwargs)

        workflow_name = cast(str, params[0])
        payload = cast(JSONSerializableMapping, params[1])
        options = cast(
            TriggerWorkflowOptions,
            params[2] if len(params) > 2 else TriggerWorkflowOptions(),
        )

        attributes = {
            OTelAttribute.WORKFLOW_NAME: workflow_name,
            OTelAttribute.ACTION_PAYLOAD: json.dumps(payload, default=str),
            OTelAttribute.PARENT_ID: options.parent_id,
            OTelAttribute.PARENT_STEP_RUN_ID: options.parent_step_run_id,
            OTelAttribute.CHILD_INDEX: options.child_index,
            OTelAttribute.CHILD_KEY: options.child_key,
            OTelAttribute.NAMESPACE: options.namespace,
            OTelAttribute.ADDITIONAL_METADATA: json.dumps(
                options.additional_metadata, default=str
            ),
            OTelAttribute.PRIORITY: options.priority,
            OTelAttribute.DESIRED_WORKER_ID: options.desired_worker_id,
            OTelAttribute.STICKY: options.sticky,
            OTelAttribute.KEY: options.key,
        }

        with self._tracer.start_as_current_span(
            "hatchet.run_workflow",
            attributes={
                f"hatchet.{k.value}": v
                for k, v in attributes.items()
                if v
                and k not in self.config.otel.excluded_attributes
                and v != "{}"
                and v != "[]"
            },
            kind=SpanKind.PRODUCER,
        ):
            options = TriggerWorkflowOptions(
                **options.model_dump(exclude={"additional_metadata"}),
                additional_metadata=_inject_traceparent_into_metadata(
                    options.additional_metadata,
                ),
            )

            return await wrapped(workflow_name, payload, options)

    def _ts_to_iso(self, ts: datetime | timestamp_pb2.Timestamp) -> str:
        if isinstance(ts, datetime):
            return ts.isoformat()
        if isinstance(ts, timestamp_pb2.Timestamp):
            return ts.ToJsonString()
        raise TypeError(f"Unsupported type for timestamp conversion: {type(ts)}")

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_schedule_workflow(
        self,
        wrapped: Callable[
            [
                str,
                list[datetime | timestamp_pb2.Timestamp],
                JSONSerializableMapping,
                ScheduleTriggerWorkflowOptions,
            ],
            v0_workflow_protos.WorkflowVersion,
        ],
        instance: AdminClient,
        args: tuple[
            str,
            list[datetime | timestamp_pb2.Timestamp],
            JSONSerializableMapping,
            ScheduleTriggerWorkflowOptions,
        ],
        kwargs: dict[
            str,
            str
            | list[datetime | timestamp_pb2.Timestamp]
            | JSONSerializableMapping
            | ScheduleTriggerWorkflowOptions,
        ],
    ) -> v0_workflow_protos.WorkflowVersion:
        params = self.extract_bound_args(wrapped, args, kwargs)

        workflow_name = cast(str, params[0])
        schedules = cast(list[datetime | timestamp_pb2.Timestamp], params[1])
        input = cast(JSONSerializableMapping, params[2])
        options = cast(
            ScheduleTriggerWorkflowOptions,
            params[3] if len(params) > 3 else ScheduleTriggerWorkflowOptions(),
        )

        attributes = {
            OTelAttribute.WORKFLOW_NAME: workflow_name,
            OTelAttribute.RUN_AT_TIMESTAMPS: json.dumps(
                [self._ts_to_iso(ts) for ts in schedules]
            ),
            OTelAttribute.ACTION_PAYLOAD: json.dumps(input, default=str),
            OTelAttribute.PARENT_ID: options.parent_id,
            OTelAttribute.PARENT_STEP_RUN_ID: options.parent_step_run_id,
            OTelAttribute.CHILD_INDEX: options.child_index,
            OTelAttribute.CHILD_KEY: options.child_key,
            OTelAttribute.NAMESPACE: options.namespace,
            OTelAttribute.ADDITIONAL_METADATA: json.dumps(
                options.additional_metadata, default=str
            ),
            OTelAttribute.PRIORITY: options.priority,
        }

        with self._tracer.start_as_current_span(
            "hatchet.schedule_workflow",
            attributes={
                f"hatchet.{k.value}": v
                for k, v in attributes.items()
                if v
                and k not in self.config.otel.excluded_attributes
                and v != "{}"
                and v != "[]"
            },
            kind=SpanKind.PRODUCER,
        ):
            options = ScheduleTriggerWorkflowOptions(
                **options.model_dump(exclude={"additional_metadata"}),
                additional_metadata=_inject_traceparent_into_metadata(
                    options.additional_metadata,
                ),
            )

            return wrapped(workflow_name, schedules, input, options)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_run_workflows(
        self,
        wrapped: Callable[
            [list[WorkflowRunTriggerConfig]],
            list[WorkflowRunRef],
        ],
        instance: AdminClient,
        args: tuple[list[WorkflowRunTriggerConfig],],
        kwargs: dict[str, list[WorkflowRunTriggerConfig]],
    ) -> list[WorkflowRunRef]:
        params = self.extract_bound_args(wrapped, args, kwargs)
        workflow_run_configs = cast(list[WorkflowRunTriggerConfig], params[0])

        num_workflows = len(workflow_run_configs)
        unique_workflow_names = {
            config.workflow_name for config in workflow_run_configs
        }

        with self._tracer.start_as_current_span(
            "hatchet.run_workflows",
            attributes={
                "hatchet.num_workflows": num_workflows,
                "hatchet.unique_workflow_names": json.dumps(
                    unique_workflow_names, default=str
                ),
            },
            kind=SpanKind.PRODUCER,
        ):
            workflow_run_configs_with_meta = [
                WorkflowRunTriggerConfig(
                    **config.model_dump(exclude={"options"}),
                    options=TriggerWorkflowOptions(
                        **config.options.model_dump(exclude={"additional_metadata"}),
                        additional_metadata=_inject_traceparent_into_metadata(
                            config.options.additional_metadata,
                        ),
                    ),
                )
                for config in workflow_run_configs
            ]

            return wrapped(workflow_run_configs_with_meta)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_async_run_workflows(
        self,
        wrapped: Callable[
            [list[WorkflowRunTriggerConfig]],
            Coroutine[None, None, list[WorkflowRunRef]],
        ],
        instance: AdminClient,
        args: tuple[list[WorkflowRunTriggerConfig],],
        kwargs: dict[str, list[WorkflowRunTriggerConfig]],
    ) -> list[WorkflowRunRef]:
        params = self.extract_bound_args(wrapped, args, kwargs)
        workflow_run_configs = cast(list[WorkflowRunTriggerConfig], params[0])
        num_workflows = len(workflow_run_configs)
        unique_workflow_names = {
            config.workflow_name for config in workflow_run_configs
        }

        with self._tracer.start_as_current_span(
            "hatchet.run_workflows",
            attributes={
                "hatchet.num_workflows": num_workflows,
                "hatchet.unique_workflow_names": json.dumps(
                    unique_workflow_names, default=str
                ),
            },
            kind=SpanKind.PRODUCER,
        ):
            workflow_run_configs_with_meta = [
                WorkflowRunTriggerConfig(
                    **config.model_dump(exclude={"options"}),
                    options=TriggerWorkflowOptions(
                        **config.options.model_dump(exclude={"additional_metadata"}),
                        additional_metadata=_inject_traceparent_into_metadata(
                            config.options.additional_metadata,
                        ),
                    ),
                )
                for config in workflow_run_configs
            ]

            return await wrapped(workflow_run_configs_with_meta)

    def _uninstrument(self, **kwargs: InstrumentKwargs) -> None:
        self.tracer_provider = NoOpTracerProvider()
        self.meter_provider = NoOpMeterProvider()

        unwrap(hatchet_sdk, "worker.runner.runner.Runner.handle_start_step_run")
        unwrap(hatchet_sdk, "worker.runner.runner.Runner.handle_cancel_action")
        unwrap(hatchet_sdk, "clients.events.EventClient.push")
        unwrap(hatchet_sdk, "clients.events.EventClient.bulk_push")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.run_workflow")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.aio_run_workflow")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.run_workflows")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.aio_run_workflows")



from enum import Enum


class OTelAttribute(str, Enum):
    ## Shared
    NAMESPACE = "namespace"
    ADDITIONAL_METADATA = "additional_metadata"
    WORKFLOW_NAME = "workflow_name"

    PRIORITY = "priority"

    ## Unfortunately named - this corresponds to all types of payloads, not just actions
    ACTION_PAYLOAD = "payload"

    ## Action
    ACTION_NAME = "action_name"
    CHILD_WORKFLOW_INDEX = "child_workflow_index"
    CHILD_WORKFLOW_KEY = "child_workflow_key"
    PARENT_WORKFLOW_RUN_ID = "parent_workflow_run_id"
    RETRY_COUNT = "retry_count"
    STEP_ID = "step_id"
    STEP_RUN_ID = "step_run_id"
    TENANT_ID = "tenant_id"
    WORKER_ID = "worker_id"
    WORKFLOW_ID = "workflow_id"
    WORKFLOW_RUN_ID = "workflow_run_id"
    WORKFLOW_VERSION_ID = "workflow_version_id"

    ## Push Event
    EVENT_KEY = "event_key"
    FILTER_SCOPE = "scope"

    ## Trigger Workflow
    PARENT_ID = "parent_id"
    PARENT_STEP_RUN_ID = "parent_step_run_id"
    CHILD_INDEX = "child_index"
    CHILD_KEY = "child_key"
    DESIRED_WORKER_ID = "desired_worker_id"
    STICKY = "sticky"
    KEY = "key"

    ## Schedule Workflow
    RUN_AT_TIMESTAMPS = "run_at_timestamps"



# coding: utf-8

"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self

from hatchet_sdk.clients.rest.models.v1_task_event import V1TaskEvent
from hatchet_sdk.clients.rest.models.v1_task_summary import V1TaskSummary
from hatchet_sdk.clients.rest.models.v1_workflow_run import V1WorkflowRun
from hatchet_sdk.clients.rest.models.workflow_run_shape_item_for_workflow_run_details import (
    WorkflowRunShapeItemForWorkflowRunDetails,
)


class V1WorkflowRunDetails(BaseModel):
    """
    V1WorkflowRunDetails
    """  # noqa: E501

    run: V1WorkflowRun
    task_events: List[V1TaskEvent] = Field(
        description="The list of task events for the workflow run", alias="taskEvents"
    )
    shape: List[WorkflowRunShapeItemForWorkflowRunDetails]
    tasks: List[V1TaskSummary]
    workflow_config: Optional[Dict[str, Any]] = Field(
        default=None, alias="workflowConfig"
    )
    __properties: ClassVar[List[str]] = [
        "run",
        "taskEvents",
        "shape",
        "tasks",
        "workflowConfig",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of V1WorkflowRunDetails from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of run
        if self.run:
            _dict["run"] = self.run.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in task_events (list)
        _items = []
        if self.task_events:
            for _item_task_events in self.task_events:
                if _item_task_events:
                    _items.append(_item_task_events.to_dict())
            _dict["taskEvents"] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in shape (list)
        _items = []
        if self.shape:
            for _item_shape in self.shape:
                if _item_shape:
                    _items.append(_item_shape.to_dict())
            _dict["shape"] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in tasks (list)
        _items = []
        if self.tasks:
            for _item_tasks in self.tasks:
                if _item_tasks:
                    _items.append(_item_tasks.to_dict())
            _dict["tasks"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of V1WorkflowRunDetails from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "run": (
                    V1WorkflowRun.from_dict(obj["run"])
                    if obj.get("run") is not None
                    else None
                ),
                "taskEvents": (
                    [V1TaskEvent.from_dict(_item) for _item in obj["taskEvents"]]
                    if obj.get("taskEvents") is not None
                    else None
                ),
                "shape": (
                    [
                        WorkflowRunShapeItemForWorkflowRunDetails.from_dict(_item)
                        for _item in obj["shape"]
                    ]
                    if obj.get("shape") is not None
                    else None
                ),
                "tasks": (
                    [V1TaskSummary.from_dict(_item) for _item in obj["tasks"]]
                    if obj.get("tasks") is not None
                    else None
                ),
                "workflowConfig": obj.get("workflowConfig"),
            }
        )
        return _obj



import asyncio
import datetime

from hatchet_sdk.clients.rest.api.workflow_api import WorkflowApi
from hatchet_sdk.clients.rest.api.workflow_run_api import WorkflowRunApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.schedule_workflow_run_request import (
    ScheduleWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_run_status import ScheduledRunStatus
from hatchet_sdk.clients.rest.models.scheduled_workflows import ScheduledWorkflows
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_filter import (
    ScheduledWorkflowsBulkDeleteFilter,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_request import (
    ScheduledWorkflowsBulkDeleteRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_delete_response import (
    ScheduledWorkflowsBulkDeleteResponse,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_item import (
    ScheduledWorkflowsBulkUpdateItem,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_request import (
    ScheduledWorkflowsBulkUpdateRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_bulk_update_response import (
    ScheduledWorkflowsBulkUpdateResponse,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_list import (
    ScheduledWorkflowsList,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_order_by_field import (
    ScheduledWorkflowsOrderByField,
)
from hatchet_sdk.clients.rest.models.update_scheduled_workflow_run_request import (
    UpdateScheduledWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.workflow_run_order_by_direction import (
    WorkflowRunOrderByDirection,
)
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import (
    BaseRestClient,
    maybe_additional_metadata_to_kv,
)
from hatchet_sdk.utils.typing import JSONSerializableMapping


class ScheduledClient(BaseRestClient):
    """
    The scheduled client is a client for managing scheduled workflows within Hatchet.
    """

    def _wra(self, client: ApiClient) -> WorkflowRunApi:
        return WorkflowRunApi(client)

    def _wa(self, client: ApiClient) -> WorkflowApi:
        return WorkflowApi(client)

    def create(
        self,
        workflow_name: str,
        trigger_at: datetime.datetime,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
    ) -> ScheduledWorkflows:
        """
        Creates a new scheduled workflow run.

        IMPORTANT: It's preferable to use `Workflow.run` (and similar) to trigger workflows if possible. This method is intended to be an escape hatch. For more details, see [the documentation](https://docs.hatchet.run/sdks/python/runnables#workflow).

        :param workflow_name: The name of the workflow to schedule.
        :param trigger_at: The datetime when the run should be triggered.
        :param input: The input data for the scheduled workflow.
        :param additional_metadata: Additional metadata associated with the future run as a key-value pair.

        :return: The created scheduled workflow instance.
        """
        with self.client() as client:
            return self._wra(client).scheduled_workflow_run_create(
                tenant=self.client_config.tenant_id,
                workflow=self.client_config.apply_namespace(workflow_name),
                schedule_workflow_run_request=ScheduleWorkflowRunRequest(
                    triggerAt=trigger_at,
                    input=input,
                    additionalMetadata=additional_metadata,
                ),
            )

    async def aio_create(
        self,
        workflow_name: str,
        trigger_at: datetime.datetime,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
    ) -> ScheduledWorkflows:
        """
        Creates a new scheduled workflow run.

        IMPORTANT: It's preferable to use `Workflow.run` (and similar) to trigger workflows if possible. This method is intended to be an escape hatch. For more details, see [the documentation](https://docs.hatchet.run/sdks/python/runnables#workflow).

        :param workflow_name: The name of the workflow to schedule.
        :param trigger_at: The datetime when the run should be triggered.
        :param input: The input data for the scheduled workflow.
        :param additional_metadata: Additional metadata associated with the future run as a key-value pair.

        :return: The created scheduled workflow instance.
        """

        return await asyncio.to_thread(
            self.create,
            workflow_name,
            trigger_at,
            input,
            additional_metadata,
        )

    def delete(self, scheduled_id: str) -> None:
        """
        Deletes a scheduled workflow run by its ID.

        :param scheduled_id: The ID of the scheduled workflow run to delete.
        :return: None
        """
        with self.client() as client:
            self._wa(client).workflow_scheduled_delete(
                tenant=self.client_config.tenant_id,
                scheduled_workflow_run=scheduled_id,
            )

    async def aio_delete(self, scheduled_id: str) -> None:
        """
        Deletes a scheduled workflow run by its ID.

        :param scheduled_id: The ID of the scheduled workflow run to delete.
        :return: None
        """
        await asyncio.to_thread(self.delete, scheduled_id)

    def update(
        self,
        scheduled_id: str,
        trigger_at: datetime.datetime,
    ) -> ScheduledWorkflows:
        """
        Reschedule a scheduled workflow run by its ID.

        Note: the server may reject rescheduling if the scheduled run has already
        triggered, or if it was created via code definition (not via API).

        :param scheduled_id: The ID of the scheduled workflow run to reschedule.
        :param trigger_at: The datetime when the run should be triggered.
        :return: The updated scheduled workflow instance.
        """
        with self.client() as client:
            return self._wa(client).workflow_scheduled_update(
                tenant=self.client_config.tenant_id,
                scheduled_workflow_run=scheduled_id,
                update_scheduled_workflow_run_request=UpdateScheduledWorkflowRunRequest(
                    triggerAt=trigger_at
                ),
            )

    async def aio_update(
        self,
        scheduled_id: str,
        trigger_at: datetime.datetime,
    ) -> ScheduledWorkflows:
        """
        Reschedule a scheduled workflow run by its ID.

        :param scheduled_id: The ID of the scheduled workflow run to reschedule.
        :param trigger_at: The datetime when the run should be triggered.
        :return: The updated scheduled workflow instance.
        """
        return await asyncio.to_thread(self.update, scheduled_id, trigger_at)

    def bulk_delete(
        self,
        *,
        scheduled_ids: list[str] | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        parent_step_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
    ) -> ScheduledWorkflowsBulkDeleteResponse:
        """
        Bulk delete scheduled workflow runs.

        Provide either:
        - `scheduled_ids` (explicit list of scheduled run IDs), or
        - one or more filter fields (`workflow_id`, `parent_workflow_run_id`, `parent_step_run_id`,
          `statuses`, `additional_metadata`)

        :param scheduled_ids: Explicit list of scheduled workflow run IDs to delete.
        :param workflow_id: Filter by workflow ID.
        :param parent_workflow_run_id: Filter by parent workflow run ID.
        :param parent_step_run_id: Filter by parent step run ID.
        :param statuses: Filter by scheduled run statuses.
        :param additional_metadata: Filter by additional metadata key/value pairs.
        :return: The bulk delete response containing deleted IDs and per-item errors.
        :raises ValueError: If neither `scheduled_ids` nor any filter field is provided.
        """
        has_filter = any(
            v is not None
            for v in (
                workflow_id,
                parent_workflow_run_id,
                parent_step_run_id,
                statuses,
                additional_metadata,
            )
        )

        if not scheduled_ids and not has_filter:
            raise ValueError(
                "bulk_delete requires either scheduled_ids or at least one filter field."
            )

        filter_obj = None
        if has_filter:
            filter_obj = ScheduledWorkflowsBulkDeleteFilter(
                workflowId=workflow_id,
                parentWorkflowRunId=parent_workflow_run_id,
                parentStepRunId=parent_step_run_id,
                additionalMetadata=maybe_additional_metadata_to_kv(additional_metadata),
                statuses=statuses,
            )

        with self.client() as client:
            return self._wa(client).workflow_scheduled_bulk_delete(
                tenant=self.client_config.tenant_id,
                scheduled_workflows_bulk_delete_request=ScheduledWorkflowsBulkDeleteRequest(
                    scheduledWorkflowRunIds=scheduled_ids,
                    filter=filter_obj,
                ),
            )

    async def aio_bulk_delete(
        self,
        *,
        scheduled_ids: list[str] | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        parent_step_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
    ) -> ScheduledWorkflowsBulkDeleteResponse:
        """
        Bulk delete scheduled workflow runs.

        :param scheduled_ids: Explicit list of scheduled workflow run IDs to delete.
        :param workflow_id: Filter by workflow ID.
        :param parent_workflow_run_id: Filter by parent workflow run ID.
        :param parent_step_run_id: Filter by parent step run ID.
        :param statuses: Filter by scheduled run statuses.
        :param additional_metadata: Filter by additional metadata key/value pairs.
        :return: The bulk delete response containing deleted IDs and per-item errors.
        :raises ValueError: If neither `scheduled_ids` nor any filter field is provided.
        """
        has_filter = any(
            v is not None
            for v in (
                workflow_id,
                parent_workflow_run_id,
                parent_step_run_id,
                statuses,
                additional_metadata,
            )
        )

        if not scheduled_ids and not has_filter:
            raise ValueError(
                "bulk_delete requires either scheduled_ids or at least one filter field."
            )

        return await asyncio.to_thread(
            self.bulk_delete,
            scheduled_ids=scheduled_ids,
            workflow_id=workflow_id,
            parent_workflow_run_id=parent_workflow_run_id,
            parent_step_run_id=parent_step_run_id,
            statuses=statuses,
            additional_metadata=additional_metadata,
        )

    def bulk_update(
        self,
        updates: (
            list[ScheduledWorkflowsBulkUpdateItem] | list[tuple[str, datetime.datetime]]
        ),
    ) -> ScheduledWorkflowsBulkUpdateResponse:
        """
        Bulk reschedule scheduled workflow runs.

        :param updates: Either:
          - a list of `(scheduled_id, trigger_at)` tuples, or
          - a list of `ScheduledWorkflowsBulkUpdateItem` objects
        :return: The bulk update response containing updated IDs and per-item errors.
        """
        update_items: list[ScheduledWorkflowsBulkUpdateItem] = []
        for u in updates:
            if isinstance(u, ScheduledWorkflowsBulkUpdateItem):
                update_items.append(u)
            else:
                scheduled_id, trigger_at = u
                update_items.append(
                    ScheduledWorkflowsBulkUpdateItem(
                        id=scheduled_id, triggerAt=trigger_at
                    )
                )

        with self.client() as client:
            return self._wa(client).workflow_scheduled_bulk_update(
                tenant=self.client_config.tenant_id,
                scheduled_workflows_bulk_update_request=ScheduledWorkflowsBulkUpdateRequest(
                    updates=update_items
                ),
            )

    async def aio_bulk_update(
        self,
        updates: (
            list[ScheduledWorkflowsBulkUpdateItem] | list[tuple[str, datetime.datetime]]
        ),
    ) -> ScheduledWorkflowsBulkUpdateResponse:
        """
        Bulk reschedule scheduled workflow runs.

        See `bulk_update` for parameter details.
        """
        return await asyncio.to_thread(self.bulk_update, updates)

    async def aio_list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
        order_by_field: ScheduledWorkflowsOrderByField | None = None,
        order_by_direction: WorkflowRunOrderByDirection | None = None,
    ) -> ScheduledWorkflowsList:
        """
        Retrieves a list of scheduled workflows based on provided filters.

        :param offset: The offset to use in pagination.
        :param limit: The maximum number of scheduled workflows to return.
        :param workflow_id: The ID of the workflow to filter by.
        :param parent_workflow_run_id: The ID of the parent workflow run to filter by.
        :param statuses: A list of statuses to filter by.
        :param additional_metadata: Additional metadata to filter by.
        :param order_by_field: The field to order the results by.
        :param order_by_direction: The direction to order the results by.

        :return: A list of scheduled workflows matching the provided filters.
        """
        return await asyncio.to_thread(
            self.list,
            offset=offset,
            limit=limit,
            workflow_id=workflow_id,
            additional_metadata=additional_metadata,
            order_by_field=order_by_field,
            order_by_direction=order_by_direction,
            parent_workflow_run_id=parent_workflow_run_id,
            statuses=statuses,
        )

    def list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: JSONSerializableMapping | None = None,
        order_by_field: ScheduledWorkflowsOrderByField | None = None,
        order_by_direction: WorkflowRunOrderByDirection | None = None,
    ) -> ScheduledWorkflowsList:
        """
        Retrieves a list of scheduled workflows based on provided filters.

        :param offset: The offset to use in pagination.
        :param limit: The maximum number of scheduled workflows to return.
        :param workflow_id: The ID of the workflow to filter by.
        :param parent_workflow_run_id: The ID of the parent workflow run to filter by.
        :param statuses: A list of statuses to filter by.
        :param additional_metadata: Additional metadata to filter by.
        :param order_by_field: The field to order the results by.
        :param order_by_direction: The direction to order the results by.

        :return: A list of scheduled workflows matching the provided filters.
        """
        with self.client() as client:
            workflow_scheduled_list = tenacity_retry(
                self._wa(client).workflow_scheduled_list, self.client_config.tenacity
            )
            return workflow_scheduled_list(
                tenant=self.client_config.tenant_id,
                offset=offset,
                limit=limit,
                order_by_field=order_by_field,
                order_by_direction=order_by_direction,
                workflow_id=workflow_id,
                additional_metadata=maybe_additional_metadata_to_kv(
                    additional_metadata
                ),
                parent_workflow_run_id=parent_workflow_run_id,
                statuses=statuses,
            )

    def get(self, scheduled_id: str) -> ScheduledWorkflows:
        """
        Retrieves a specific scheduled workflow by scheduled run trigger ID.

        :param scheduled_id: The scheduled workflow trigger ID to retrieve.
        :return: The requested scheduled workflow instance.
        """

        with self.client() as client:
            workflow_scheduled_get = tenacity_retry(
                self._wa(client).workflow_scheduled_get, self.client_config.tenacity
            )
            return workflow_scheduled_get(
                tenant=self.client_config.tenant_id,
                scheduled_workflow_run=scheduled_id,
            )

    async def aio_get(self, scheduled_id: str) -> ScheduledWorkflows:
        """
        Retrieves a specific scheduled workflow by scheduled run trigger ID.

        :param scheduled_id: The scheduled workflow trigger ID to retrieve.
        :return: The requested scheduled workflow instance.
        """
        return await asyncio.to_thread(self.get, scheduled_id)



import asyncio

from hatchet_sdk.clients.rest.api.filter_api import FilterApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.v1_create_filter_request import (
    V1CreateFilterRequest,
)
from hatchet_sdk.clients.rest.models.v1_filter import V1Filter
from hatchet_sdk.clients.rest.models.v1_filter_list import V1FilterList
from hatchet_sdk.clients.rest.models.v1_update_filter_request import (
    V1UpdateFilterRequest,
)
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import BaseRestClient
from hatchet_sdk.utils.typing import JSONSerializableMapping


class FiltersClient(BaseRestClient):
    """
    The filters client is a client for interacting with Hatchet's filters API.
    """

    def _fa(self, client: ApiClient) -> FilterApi:
        return FilterApi(client)

    async def aio_list(
        self,
        limit: int | None = None,
        offset: int | None = None,
        workflow_ids: list[str] | None = None,
        scopes: list[str] | None = None,
    ) -> V1FilterList:
        """
        List filters for a given tenant.

        :param limit: The maximum number of filters to return.
        :param offset: The number of filters to skip before starting to collect the result set.
        :param workflow_ids: A list of workflow IDs to filter by.
        :param scopes: A list of scopes to filter by.

        :return: A list of filters matching the specified criteria.
        """
        return await asyncio.to_thread(self.list, limit, offset, workflow_ids, scopes)

    def list(
        self,
        limit: int | None = None,
        offset: int | None = None,
        workflow_ids: list[str] | None = None,
        scopes: list[str] | None = None,
    ) -> V1FilterList:
        """
        List filters for a given tenant.

        :param limit: The maximum number of filters to return.
        :param offset: The number of filters to skip before starting to collect the result set.
        :param workflow_ids: A list of workflow IDs to filter by.
        :param scopes: A list of scopes to filter by.

        :return: A list of filters matching the specified criteria.
        """
        with self.client() as client:
            v1_filter_list = tenacity_retry(
                self._fa(client).v1_filter_list, self.client_config.tenacity
            )
            return v1_filter_list(
                tenant=self.tenant_id,
                limit=limit,
                offset=offset,
                workflow_ids=workflow_ids,
                scopes=scopes,
            )

    def get(
        self,
        filter_id: str,
    ) -> V1Filter:
        """
        Get a filter by its ID.

        :param filter_id: The ID of the filter to retrieve.

        :return: The filter with the specified ID.
        """
        with self.client() as client:
            v1_filter_get = tenacity_retry(
                self._fa(client).v1_filter_get, self.client_config.tenacity
            )
            return v1_filter_get(
                tenant=self.tenant_id,
                v1_filter=filter_id,
            )

    async def aio_get(
        self,
        filter_id: str,
    ) -> V1Filter:
        """
        Get a filter by its ID.

        :param filter_id: The ID of the filter to retrieve.

        :return: The filter with the specified ID.
        """
        return await asyncio.to_thread(self.get, filter_id)

    def create(
        self,
        workflow_id: str,
        expression: str,
        scope: str,
        payload: JSONSerializableMapping | None = None,
    ) -> V1Filter:
        """
        Create a new filter.

        :param workflow_id: The ID of the workflow to associate with the filter.
        :param expression: The expression to evaluate for the filter.
        :param scope: The scope for the filter.
        :param payload: The payload to send with the filter.

        :return: The created filter.
        """
        with self.client() as client:
            return self._fa(client).v1_filter_create(
                tenant=self.tenant_id,
                v1_create_filter_request=V1CreateFilterRequest(
                    workflowId=workflow_id,
                    expression=expression,
                    scope=scope,
                    payload=payload,
                ),
            )

    async def aio_create(
        self,
        workflow_id: str,
        expression: str,
        scope: str,
        payload: JSONSerializableMapping | None = None,
    ) -> V1Filter:
        """
        Create a new filter.

        :param workflow_id: The ID of the workflow to associate with the filter.
        :param expression: The expression to evaluate for the filter.
        :param scope: The scope for the filter.
        :param payload: The payload to send with the filter.

        :return: The created filter.
        """
        return await asyncio.to_thread(
            self.create, workflow_id, expression, scope, payload
        )

    def delete(
        self,
        filter_id: str,
    ) -> V1Filter:
        """
        Delete a filter by its ID.

        :param filter_id: The ID of the filter to delete.
        :return: The deleted filter.
        """
        with self.client() as client:
            return self._fa(client).v1_filter_delete(
                tenant=self.tenant_id,
                v1_filter=filter_id,
            )

    async def aio_delete(
        self,
        filter_id: str,
    ) -> V1Filter:
        """
        Delete a filter by its ID.

        :param filter_id: The ID of the filter to delete.
        :return: The deleted filter.
        """
        return await asyncio.to_thread(self.delete, filter_id)

    def update(
        self,
        filter_id: str,
        updates: V1UpdateFilterRequest,
    ) -> V1Filter:
        """
        Update a filter by its ID.

        :param filter_id: The ID of the filter to delete.
        :param updates: The updates to apply to the filter.
        :return: The updated filter.
        """
        with self.client() as client:
            return self._fa(client).v1_filter_update(
                tenant=self.tenant_id,
                v1_filter=filter_id,
                v1_update_filter_request=updates,
            )

    async def aio_update(
        self,
        filter_id: str,
        updates: V1UpdateFilterRequest,
    ) -> V1Filter:
        """
        Update a filter by its ID.

        :param filter_id: The ID of the filter to delete.
        :param updates: The updates to apply to the filter.
        :return: The updated filter.
        """
        return await asyncio.to_thread(self.update, filter_id, updates)



# coding: utf-8

"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictStr
from typing_extensions import Self

from hatchet_sdk.clients.rest.models.api_meta_auth import APIMetaAuth
from hatchet_sdk.clients.rest.models.api_meta_posthog import APIMetaPosthog


class APIMeta(BaseModel):
    """
    APIMeta
    """  # noqa: E501

    auth: Optional[APIMetaAuth] = None
    pylon_app_id: Optional[StrictStr] = Field(
        default=None,
        description="the Pylon app ID for usepylon.com chat support",
        alias="pylonAppId",
    )
    posthog: Optional[APIMetaPosthog] = None
    allow_signup: Optional[StrictBool] = Field(
        default=None,
        description="whether or not users can sign up for this instance",
        alias="allowSignup",
    )
    allow_invites: Optional[StrictBool] = Field(
        default=None,
        description="whether or not users can invite other users to this instance",
        alias="allowInvites",
    )
    allow_create_tenant: Optional[StrictBool] = Field(
        default=None,
        description="whether or not users can create new tenants",
        alias="allowCreateTenant",
    )
    allow_change_password: Optional[StrictBool] = Field(
        default=None,
        description="whether or not users can change their password",
        alias="allowChangePassword",
    )
    __properties: ClassVar[List[str]] = [
        "auth",
        "pylonAppId",
        "posthog",
        "allowSignup",
        "allowInvites",
        "allowCreateTenant",
        "allowChangePassword",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of APIMeta from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of auth
        if self.auth:
            _dict["auth"] = self.auth.to_dict()
        # override the default output from pydantic by calling `to_dict()` of posthog
        if self.posthog:
            _dict["posthog"] = self.posthog.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of APIMeta from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "auth": (
                    APIMetaAuth.from_dict(obj["auth"])
                    if obj.get("auth") is not None
                    else None
                ),
                "pylonAppId": obj.get("pylonAppId"),
                "posthog": (
                    APIMetaPosthog.from_dict(obj["posthog"])
                    if obj.get("posthog") is not None
                    else None
                ),
                "allowSignup": obj.get("allowSignup"),
                "allowInvites": obj.get("allowInvites"),
                "allowCreateTenant": obj.get("allowCreateTenant"),
                "allowChangePassword": obj.get("allowChangePassword"),
            }
        )
        return _obj
